"""
.. module:: samplers.mcmc.proposal

:Synopsis: proposal distributions
:Author: Antony Lewis and Jesus Torrado

Using the covariance matrix to give the proposal directions typically
significantly increases the acceptance rate and gives faster movement
around parameter space.

We generate a random basis in the eigenvectors, then cycle through them
proposing changes to each, then generate a new random basis.
The distance proposal in the random direction is given by a two-D Gaussian
radial function mixed with an exponential, which is quite robust to wrong width estimates

See https://arxiv.org/abs/1304.4473
"""

import logging
from itertools import chain

import numpy as np

from cobaya.log import LoggedError, HasLogger
from cobaya.tools import choleskyL


class IndexCycler:
    def __init__(self, n, random_state):
        self.n = n
        self.loop_index = -1
        self.random_state = random_state


class CyclicIndexRandomizer(IndexCycler):
    def __init__(self, n, random_state):
        if isinstance(n, int):
            self.sorted_indices = list(range(n))
        else:
            self.sorted_indices = n
            n = len(n)
        super().__init__(n, random_state)
        if n <= 2:
            self.indices = list(range(n))

    def next(self):
        """
        Get the next random index, or alternate for two or less.

        :return: index
        """
        self.loop_index = (self.loop_index + 1) % self.n
        if self.loop_index == 0 and self.n > 2:
            self.indices = self.random_state.permutation(self.sorted_indices)
        return self.indices[self.loop_index]


try:
    import numba
except (ImportError, SystemError):
    # SystemError caused usually by incompatible numpy version
    from scipy.stats import special_ortho_group

    random_SO_N = special_ortho_group.rvs
    numba = None
else:
    import warnings

    def random_SO_N(dim, random_state):
        """
        Draw random samples from SO(N).
        Equivalent to scipy function but about 10x faster
        Parameters
        ----------
        dim : integer
            Dimension of rotation space (N).
        random_state: generator
        Returns
        -------
        rvs : Random size N-dimensional matrices, dimension (dim, dim)

        """
        dim = np.int64(dim)
        H = np.eye(dim)
        xx = random_state.standard_normal(size=(dim + 2) * (dim - 1) // 2)
        _rvs(dim, xx, H)
        return H

    logging.getLogger('numba').setLevel(logging.ERROR)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")

        @numba.njit("void(int64,float64[::1],float64[:,::1])")
        def _rvs(dim, xx, H):
            D = np.empty((dim,))
            ix = 0
            for n in range(dim - 1):
                x = xx[ix:ix + dim - n]
                ix += dim - n
                norm2 = np.dot(x, x)
                x0 = x[0].item()
                D[n] = np.sign(x[0]) if x[0] != 0 else 1
                x[0] += D[n] * np.sqrt(norm2)
                x /= np.sqrt((norm2 - x0 ** 2 + x[0] ** 2) / 2.)
                # Householder transformation
                tmp = np.dot(H[:, n:], x)
                H[:, n:] -= np.outer(tmp, x)
            D[-1] = (-1) ** (dim - 1) * D[:-1].prod()
            H[:, :] = (D * H.T).T


class RandDirectionProposer(IndexCycler):
    def propose_vec(self, scale: float = 1):
        """
        Propose a random n-dimension vector for n>1

        :param scale: units for the distance
        :return: array with vector
        """
        self.loop_index = (self.loop_index + 1) % self.n
        if self.loop_index == 0:
            self.R = random_SO_N(self.n, random_state=self.random_state)
        return self.R[:, self.loop_index] * self.propose_r() * scale

    def propose_r(self):
        """
        Radial proposal. By default a mixture of an exponential and 2D Gaussian radial
        proposal (to make wider tails and more mass near zero, so more robust to scale
        misestimation)

        :return: random distance (unit scale)
        """
        if self.random_state.uniform() < 0.33:
            return self.random_state.standard_exponential()
        else:
            return np.sqrt(self.random_state.chisquare(min(self.n, 2)))


class RandProposer1D(RandDirectionProposer):
    def propose_vec(self, scale: float = 1):
        return np.array([self.propose_r() * scale if self.random_state.integers(2)
                         else -self.propose_r() * scale])


class BlockedProposer(HasLogger):
    def __init__(self, parameter_blocks, random_state, oversampling_factors=None,
                 i_last_slow_block=None, proposal_scale=2.4):
        """
        Proposal density for fast and slow parameters, where parameters are
        grouped into blocks which are changed at the same time.

        :param parameter_blocks: list of arrays of parameter indices in each block,
            with blocks sorted by ascending speed.
        :param oversampling_factors: list of *int* oversampling factors *per parameter*,
            i.e. a factor of n for a block of dimension d would mean n*d jumps for that
            block per full cycle, whereas a factor of 1 for all blocks (default) means
            that all *directions* are treated equally (but the proposals are still
            block-wise).
        :param i_last_slow_block: index of the last block considered slow.
            By default, all blocks are considered slow.
        :param proposal_scale: overall scale for the proposal.
        """
        self.set_logger(lowercase=True)
        self.random_state = random_state
        self.proposal_scale = proposal_scale
        if oversampling_factors is None:
            self.oversampling_factors = np.ones(len(parameter_blocks), dtype=int)
        else:
            if len(oversampling_factors) != len(parameter_blocks):
                raise LoggedError(
                    self.log, "List of oversampling factors has a different length that "
                              "list of blocks: %d vs %d.",
                    len(oversampling_factors), len(parameter_blocks))
            if np.any(oversampling_factors != np.floor(np.array(oversampling_factors))):
                raise LoggedError(
                    self.log, "Oversampling factors must be integer! Got %r.",
                    oversampling_factors)
            self.oversampling_factors = np.array(oversampling_factors, dtype=int)
        # Binary fast-slow split
        self.i_last_slow_block = i_last_slow_block
        if self.i_last_slow_block is None:
            self.i_last_slow_block = len(parameter_blocks) - 1
        else:
            if self.i_last_slow_block > len(parameter_blocks) - 1:
                raise LoggedError(
                    self.log,
                    "The index given for the last slow block, %d, is not valid: "
                    "there are only %d blocks.",
                    self.i_last_slow_block, len(parameter_blocks))
        n_all = sum(len(b) for b in parameter_blocks)
        n_slow = sum(len(b) for b in parameter_blocks[:1 + self.i_last_slow_block])
        self.nsamples_slow = 0
        self.nsamples_fast = 0
        if set(chain(*parameter_blocks)) != set(range(n_all)):
            raise LoggedError(self.log,
                              "The blocks do not contain all the parameter indices.")
        # Prepare indices for the cycler, repeated if there is oversampling
        self.n_block = np.array([len(b) for b in parameter_blocks])
        indices_repeated = list(chain(
            *[list(chain(*[[p] * o for p in block]))
              for block, o in zip(parameter_blocks, oversampling_factors)]))
        # Mapping between internal indices, sampler parameter indices and blocks:
        # let i=0,1,... be the indices of the parameters for the sampler,
        # and j=0,1,... be the indices of the parameters as the proposer manages them
        # i.e. if passed blocks=[[1,2],[0]] (those are the i's),
        # then the j's are [0 (for 1), 1 (for 2), 2 (for 0)].
        # iblock is the index of the blocks, which in term of j indices is simply
        # [0,0,1] in this example
        self.i_of_j = np.array(list(chain(*parameter_blocks)))
        self.iblock_of_j = list(
            chain(*[[iblock] * len(b) for iblock, b in enumerate(parameter_blocks)]))
        # Creating the blocked proposers
        self.proposer = [(RandDirectionProposer(len(b), random_state) if len(b) > 1
                          else RandProposer1D(1, random_state)) for b in parameter_blocks]
        # Starting j index of each block
        self.j_start = [len(list(chain(*parameter_blocks[:iblock])))
                        for iblock, b in enumerate(parameter_blocks)]
        # Parameter cyclers, cycling over the j's
        self.parameter_cycler = CyclicIndexRandomizer(indices_repeated, random_state)
        # These ones are used by fast dragging only
        self.parameter_cycler_slow = CyclicIndexRandomizer(n_slow, random_state)
        self.parameter_cycler_fast = CyclicIndexRandomizer(n_all - n_slow, random_state)

    def d(self):
        return len(self.i_of_j)

    def get_proposal(self, P):
        self.current_iblock = self.iblock_of_j[self.parameter_cycler.next()]
        if self.current_iblock <= self.i_last_slow_block:
            self.nsamples_slow += 1
        else:
            self.nsamples_fast += 1
        self.get_block_proposal(P, self.current_iblock)

    def get_proposal_slow(self, P):
        current_iblock_slow = self.iblock_of_j[self.parameter_cycler_slow.next()]
        self.nsamples_slow += 1
        self.get_block_proposal(P, current_iblock_slow)

    def get_proposal_fast(self, P):
        self.nsamples_fast += 1
        current_iblock_fast = self.iblock_of_j[self.parameter_cycler_slow.n +
                                               self.parameter_cycler_fast.next()]
        self.get_block_proposal(P, current_iblock_fast)

    def get_block_proposal(self, P, iblock):
        vec_standardized = self.proposer[iblock].propose_vec(self.proposal_scale)
        P[self.i_of_j[self.j_start[iblock]:]] += (self.transform[iblock]
                                                  .dot(vec_standardized))

    def set_covariance(self, propose_matrix):
        """
        Take covariance of sampled parameters (propose_matrix), and construct orthonormal
        parameters where orthonormal parameters are grouped in blocks by speed, so changes
        in the slowest block changes slow and fast parameters, but changes in the fastest
        block only changes fast parameters

        :param propose_matrix: covariance matrix for the sampled parameters.
        """
        if propose_matrix.shape[0] != self.d():
            raise LoggedError(
                self.log, "The covariance matrix does not have the correct dimension: "
                          "it's %d, but it should be %d.", propose_matrix.shape[0],
                self.d())
        if not (np.allclose(propose_matrix.T, propose_matrix) and
                np.all(np.linalg.eigvals(propose_matrix) > 0)):
            raise LoggedError(self.log, "The given covmat is not a positive-definite, "
                                        "symmetric square matrix.")
        self.propose_matrix = propose_matrix.copy()
        propose_matrix_j_sorted = self.propose_matrix[np.ix_(self.i_of_j, self.i_of_j)]
        sigmas_diag, L = choleskyL(propose_matrix_j_sorted, return_scale_free=True)
        # Store the basis as transformation matrices
        self.transform = []
        for j_start, bp in zip(self.j_start, self.proposer):
            j_end = j_start + bp.n
            self.transform += [sigmas_diag[j_start:, j_start:].dot(L[j_start:,
                                                                   j_start:j_end])]

    def get_covariance(self):
        return self.propose_matrix.copy()

    def get_scale(self):
        return self.proposal_scale



from cobaya.likelihoods.base_classes import DES


class galaxy_galaxy(DES):
    r"""
    Galaxy-galaxy lensing data from the first year of the Dark Energy Survey (DES Y1)
    \cite{Abbott:2017wau}.
    """
    bibtex_file = 'des_y1.bibtex'



# Local
from cobaya.likelihood import Likelihood


class _test(Likelihood):
    """
    Likelihood that evaluates to 1.
    """

    def logp(self, **params_values):
        self.wait()
        params_values["_derived"]["b1"] = 0
        return 0.



#!/usr/bin/env python

import argparse
import os
from . import jobqueue


def run_single():
    parser = argparse.ArgumentParser(
        prog="cobaya run-job",
        description="Submit a single job to queue")

    parser.add_argument('input_file', nargs='+')

    jobqueue.addArguments(parser)

    args = parser.parse_args()

    ini = [ini.replace('.ini', '').replace('.yaml', '') for ini in args.input_file]

    jobqueue.submitJob(os.path.basename(ini[0]), ini, msg=True, **args.__dict__)


if __name__ == "__main__":
    run_single()



from cobaya.likelihoods.base_classes import Planck2018Clik


class TTTEEE(Planck2018Clik):
    r"""
    High-$\ell$ temperature and polarization \textsc{plik} likelihood of Planck's 2018
    data release \cite{Aghanim:2019ame}.
    """
    pass



"""
.. module:: sampler

:Synopsis: Base class for samplers and other parameter-space explorers
:Author: Jesus Torrado

cobaya includes by default a
:doc:`Monte Carlo Markov Chain (MCMC) sampler <sampler_mcmc>`
(a direct translation from `CosmoMC <https://cosmologist.info/cosmomc/>`_) and a dummy
:doc:`evaluate <sampler_evaluate>` sampler that simply evaluates the posterior at a given
(or sampled) reference point. It also includes an interface to the
:doc:`PolyChord sampler <sampler_polychord>` (needs to be installed separately).

The sampler to use is specified by a `sampler` block in the input file, whose only member
is the sampler used, containing some options, if necessary.

.. code-block:: yaml

   sampler:
     mcmc:
       max_samples: 1000

or

.. code-block:: yaml

   sampler:
     polychord:
       path: /path/to/cosmo/PolyChord

Samplers can in general be swapped in the input file without needing to modify any other
block of the input.

In the cobaya code tree, each sampler is placed in its own folder, containing a file
defining the sampler's class, which inherits from the :class:`cobaya.Sampler`, and a
``[sampler_name].yaml`` file, containing all possible user-specified options for the
sampler and their default values. Whatever option is defined in this file automatically
becomes an attribute of the sampler's instance.

To implement your own sampler, or an interface to an external one, simply create a folder
under the ``cobaya/cobaya/samplers/`` folder and include the two files described above.
Your class needs to inherit from the :class:`cobaya.Sampler` class below, and needs to
implement only the methods ``initialize``, ``_run``, and ``products``.

"""
# Global
import os
import numpy as np
from typing import Optional, Sequence, Mapping, Union
from itertools import chain
from numpy.random import SeedSequence, default_rng

# Local
from cobaya.conventions import Extension, packages_path_input
from cobaya.typing import InfoDict, SamplersDict, SamplerDict
from cobaya.tools import deepcopy_where_possible, find_with_regexp, recursive_update, \
    str_to_list
from cobaya.model import Model
from cobaya.log import LoggedError, get_logger, is_debug
from cobaya.yaml import yaml_load_file, yaml_dump
from cobaya.component import CobayaComponent, get_component_class
from cobaya.input import update_info, is_equal_info, get_preferred_old_values
from cobaya.output import OutputDummy, Output
from cobaya import mpi


def get_sampler_name_and_class(info_sampler: SamplersDict, logger=None):
    """
    Auxiliary function to retrieve the class of the required sampler.
    """
    check_sane_info_sampler(info_sampler)
    name = list(info_sampler)[0]
    sampler_class = get_component_class(name, kind="sampler", logger=logger)
    assert issubclass(sampler_class, Sampler)
    return name, sampler_class


def check_sane_info_sampler(info_sampler: SamplersDict):
    if not info_sampler:
        raise LoggedError(__name__, "No sampler given!")
    try:
        list(info_sampler)[0]
    except AttributeError:
        raise LoggedError(
            __name__, "The sampler block must be a dictionary 'sampler: {options}'.")
    if len(info_sampler) > 1:
        raise LoggedError(__name__, "Only one sampler currently supported at a time.")


def check_sampler_info(info_old: Optional[SamplersDict],
                       info_new: SamplersDict, is_resuming=False):
    """
    Checks compatibility between the new sampler info and that of a pre-existing run.

    Done separately from `Output.check_compatible_and_dump` because there may be
    multiple samplers mentioned in an `updated.yaml` file, e.g. `MCMC` + `Minimize`.
    """
    logger_sampler = get_logger(__name__)
    if not info_old:
        return
    # TODO: restore this at some point: just append minimize info to the old one
    # There is old info, but the new one is Minimizer and the old one is not
    # if (len(info_old) == 1 and list(info_old) != ["minimize"] and
    #      list(info_new) == ["minimize"]):
    #     # In-place append of old+new --> new
    #     aux = info_new.pop("minimize")
    #     info_new.update(info_old)
    #     info_new.update({"minimize": aux})
    #     info_old = {}
    #     keep_old = {}
    if list(info_old) != list(info_new) and list(info_new) == ["minimize"]:
        return
    if list(info_old) == list(info_new):
        # Restore some selected old values for some classes
        keep_old = get_preferred_old_values({"sampler": info_old})
        info_new = recursive_update(info_new, keep_old.get("sampler", {}))
    if not is_equal_info(
            {"sampler": info_old}, {"sampler": info_new}, strict=False):
        if is_resuming:
            raise LoggedError(
                logger_sampler, "Old and new Sampler information not compatible! "
                                "Resuming not possible!")
        else:
            raise LoggedError(
                logger_sampler, "Found old Sampler information which is not compatible "
                                "with the new one. Delete the previous output manually, "
                                "or automatically with either "
                                "'-f', '--force', 'force: True'")


def get_sampler(info_sampler: SamplersDict, model: Model, output: Optional[Output] = None,
                packages_path: Optional[str] = None) -> 'Sampler':
    assert isinstance(info_sampler, Mapping), (
        "The first argument must be a dictionary with the info needed for the sampler. "
        "If you were trying to pass the name of an input file instead, "
        "load it first with 'cobaya.input.load_input', "
        "or, if you were passing a yaml string, load it with 'cobaya.yaml.yaml_load'.")
    logger_sampler = get_logger(__name__)
    info_sampler = deepcopy_where_possible(info_sampler)
    if output is None:
        output = OutputDummy()
    # Check and update info
    check_sane_info_sampler(info_sampler)
    updated_info_sampler = update_info(
        {"sampler": info_sampler})["sampler"]  # type: ignore
    if is_debug(logger_sampler):
        logger_sampler.debug(
            "Input info updated with defaults (dumped to YAML):\n%s",
            yaml_dump(updated_info_sampler))
    # Get sampler class & check resume/force compatibility
    sampler_name, sampler_class = get_sampler_name_and_class(
        updated_info_sampler, logger=logger_sampler)
    check_sampler_info(
        (output.reload_updated_info(use_cache=True) or {}).get("sampler"),
        updated_info_sampler, is_resuming=output.is_resuming())
    # Check if resumable run
    sampler_class.check_force_resume(output, info=updated_info_sampler[sampler_name])
    # Instantiate the sampler
    sampler_instance = sampler_class(updated_info_sampler[sampler_name], model,
                                     output, packages_path=packages_path)
    # If output, dump updated
    if output:
        to_dump = model.info()
        to_dump["sampler"] = {sampler_name: sampler_instance.info()}
        to_dump["output"] = os.path.join(output.folder, output.prefix)
        output.check_and_dump_info(None, to_dump, check_compatible=False)
    return sampler_instance


class Sampler(CobayaComponent):
    """Base class for samplers."""

    # What you *must* implement to create your own sampler:

    seed: Union[None, int, Sequence[int]]
    version: Optional[Union[dict, str]] = None

    _rng: np.random.Generator

    def initialize(self):
        """
        Initializes the sampler: prepares the samples' collection,
        prepares the output, deals with MPI scheduling, imports an external sampler, etc.

        Options defined in the ``defaults.yaml`` file in the sampler's folder are
        automatically recognized as attributes, with the value given in the input file,
        if redefined there.

        The prior and likelihood are also accessible through the attributes with the same
        names.
        """
        pass

    def run(self):
        """
        Runs the main part of the algorithm of the sampler.
        Normally, it looks somewhat like

        .. code-block:: python

           while not [convergence criterion]:
               [do one more step]
               [update the collection of samples]
        """
        pass

    def products(self) -> InfoDict:
        """
        Returns the products expected in a scripted call of cobaya,
        (e.g. a collection of samples or a list of them).
        """
        return {}

    @property
    def random_state(self) -> np.random.Generator:
        return self._rng

    @property
    def model(self) -> Model:
        return self._model

    @property
    def output(self) -> Output:
        return self._output

    # Private methods: just ignore them:
    def __init__(self, info_sampler: SamplerDict, model: Model,
                 output=Optional[Output], packages_path: Optional[str] = None,
                 name: Optional[str] = None):
        """
        Actual initialization of the class. Loads the default and input information and
        call the custom ``initialize`` method.

        [Do not modify this one.]
        """
        self._model = model
        self._output = output
        self._updated_info = deepcopy_where_possible(info_sampler)
        super().__init__(info_sampler, packages_path=packages_path,
                         name=name, initialize=False, standalone=False)
        if not model.parameterization.sampled_params():
            self.mpi_warning("No sampled parameters requested! "
                             "This will fail for non-mock samplers.")
        # Load checkpoint info, if resuming
        if self.output.is_resuming() and not isinstance(self, Minimizer):
            checkpoint_info = None
            if mpi.is_main_process():
                try:
                    checkpoint_info = yaml_load_file(self.checkpoint_filename())

                    if self.get_name() not in checkpoint_info["sampler"]:
                        raise LoggedError(
                            self.log, "Checkpoint file found at '%s' "
                                      "but it corresponds to a different sampler.",
                            self.checkpoint_filename())
                except (IOError, TypeError):
                    pass
            checkpoint_info = mpi.share_mpi(checkpoint_info)
            if checkpoint_info:
                self.set_checkpoint_info(checkpoint_info)
                self.mpi_info("Resuming from previous sample!")
        elif not isinstance(self, Minimizer) and mpi.is_main_process():
            try:
                output.delete_file_or_folder(self.checkpoint_filename())
                output.delete_file_or_folder(self.progress_filename())
            except (OSError, TypeError):
                pass
        self._set_rng()
        self.initialize()
        model.set_cache_size(self._get_requested_cache_size())
        # Add to the updated info some values which are
        # only available after initialisation
        self._updated_info["version"] = self.get_version()

    def info(self):
        """
        Returns a copy of the information used to initialise the sampler,
        including defaults and some new values that are only available after
        initialisation.
        """
        return deepcopy_where_possible(self._updated_info)

    def checkpoint_filename(self):
        if self.output:
            return os.path.join(
                self.output.folder, self.output.prefix + Extension.checkpoint)
        return None

    def progress_filename(self):
        if self.output:
            return os.path.join(
                self.output.folder, self.output.prefix + Extension.progress)
        return None

    def set_checkpoint_info(self, checkpoint_info):
        for k, v in checkpoint_info["sampler"][self.get_name()].items():
            setattr(self, k, v)
        # check if convergence parameters changed, and if so converged=False
        old_info = self.output.reload_updated_info(use_cache=True)
        assert old_info
        if self.converge_info_changed(old_info["sampler"][self.get_name()],
                                      self._updated_info):
            self.converged = False

    def converge_info_changed(self, old_info, new_info):
        return old_info != new_info

    def _get_requested_cache_size(self):
        """
        Override this for samplers than need more than 3 states cached
        per theory/likelihood.

        :return: number of points to cache
        """
        return 3

    def _set_rng(self):
        """
        Initialize random generator stream. For seeded runs, sets the state reproducibly.
        """
        # TODO: checkpointing save of self._rng.bit_generator.state per process
        if mpi.is_main_process():
            seed = getattr(self, "seed", None)
            if seed is not None:
                self.mpi_warning("This run has been SEEDED with seed %s", seed)
            ss = SeedSequence(seed)
            child_seeds = ss.spawn(mpi.size())
        else:
            child_seeds = None
        ss = mpi.scatter(child_seeds)
        self._entropy = ss.entropy  # for debugging store for reproducibility
        self._rng = default_rng(ss)

    @classmethod
    def output_files_regexps(cls, output, info=None, minimal=False):
        """
        Returns a list of tuples `(regexp, root)` of output files potentially produced.
        If `root` in the tuple is `None`, `output.folder` is used.

        If `minimal=True`, returns regexp's for the files that should really not be there
        when we are not resuming.
        """
        return []

    @classmethod
    @mpi.root_only
    def delete_output_files(cls, output, info=None):
        if output:
            for (regexp, root) in cls.output_files_regexps(output, info=info):
                # Special case: CovmatSampler's may have been given a covmat with the same
                # name that the output one. In that case, don't delete it!
                if issubclass(cls, CovmatSampler) and info:
                    if regexp.pattern.rstrip("$").endswith(Extension.covmat):
                        covmat_file = info.get("covmat", "")
                        if (isinstance(covmat_file, str) and covmat_file ==
                                getattr(regexp.match(covmat_file), "group",
                                        lambda: None)()):
                            continue
                output.delete_with_regexp(regexp, root)

    @classmethod
    def check_force_resume(cls, output, info=None):
        """
        Performs the necessary checks on existing files if resuming or forcing
        (including deleting some output files when forcing).
        """
        if not output:
            return
        resuming: Optional[bool]
        if mpi.is_main_process():
            resuming = False
            if output.force:
                cls.delete_output_files(output, info=info)
            elif any(find_with_regexp(regexp, root or output.folder) for (regexp, root)
                     in cls.output_files_regexps(output=output, info=info, minimal=True)):
                if output.is_resuming():
                    output.log.info("Found an old sample. Resuming.")
                    resuming = True
                else:
                    raise LoggedError(
                        output.log, "Delete the previous output manually, automatically "
                                    "('-%s', '--%s', '%s: True')" % (
                                        "force"[0], "force", "force") +
                                    " or request resuming ('-%s', '--%s', '%s: True')" % (
                                        "resume"[0], "resume", "resume"))
            else:
                if output.is_resuming():
                    output.log.info(
                        "Did not find an old sample. Cleaning up and starting anew.")
                # Clean up old files, and set resuming=False,
                # regardless of requested value
                cls.delete_output_files(output, info=info)
        else:
            resuming = None
        output.set_resuming(resuming)


class Minimizer(Sampler):
    """
    base class for minimizers
    """


class CovmatSampler(Sampler):
    """
    Parent class for samplers that are initialised with a covariance matrix.
    """
    covmat_params: Sequence[str]
    # Amount by which to shrink covmat diagonals when set from priors or reference.
    fallback_covmat_scale: float = 4

    @mpi.from_root
    def _load_covmat(self, prefer_load_old, auto_params=None):
        if prefer_load_old and os.path.exists(self.covmat_filename()):
            covmat = np.atleast_2d(np.loadtxt(self.covmat_filename()))
            self.mpi_info("Covariance matrix from previous sample.")
            return covmat, []
        else:
            return self.initial_proposal_covmat(auto_params=auto_params)

    # noinspection PyUnboundLocalVariable
    def initial_proposal_covmat(self, auto_params=None):
        """
        Build the initial covariance matrix, using the data provided, in descending order
        of priority:
        1. "covmat" field in the sampler block (including `auto` search).
        2. "proposal" field for each parameter.
        3. variance of the reference pdf.
        4. variance of the prior pdf.

        The covariances between parameters when both are present in a covariance matrix
        provided through option 1 are preserved. All other covariances are assumed 0.

        If `covmat: auto`, use the keyword `auto_params` to restrict the parameters for
        which a covariance matrix is searched (default: None, meaning all sampled params).
        """
        params_infos = self.model.parameterization.sampled_params_info()
        covmat = np.diag([np.nan] * len(params_infos))
        # Try to generate it automatically
        self.covmat = getattr(self, 'covmat', None)
        if isinstance(self.covmat, str) and self.covmat.lower() == "auto":
            params_infos_covmat = deepcopy_where_possible(params_infos)
            for p in list(params_infos_covmat):
                if p not in (auto_params or []):
                    params_infos_covmat.pop(p, None)
            auto_covmat = self.model.get_auto_covmat(params_infos_covmat,
                                                     random_state=self._rng)
            if auto_covmat:
                self.covmat = os.path.join(auto_covmat["folder"], auto_covmat["name"])
                self.log.info("Covariance matrix selected automatically: %s", self.covmat)
            else:
                self.covmat = None
                self.log.info("Could not automatically find a good covmat. "
                              "Will generate from parameter info (proposal and prior).")
        # If given, load and test the covariance matrix
        loaded_params: Sequence[str]
        if isinstance(self.covmat, str):
            covmat_pre = "{%s}" % packages_path_input
            if self.covmat.startswith(covmat_pre):
                self.covmat = self.covmat.format(
                    **{packages_path_input: self.packages_path}).replace("/", os.sep)
            try:
                with open(self.covmat, "r", encoding="utf-8-sig") as file_covmat:
                    header = file_covmat.readline()
                loaded_covmat = np.loadtxt(self.covmat)
                self.log.debug(f"Loaded a covariance matrix from '{self.covmat}'")
            except TypeError:
                raise LoggedError(self.log, "The property 'covmat' must be a file name,"
                                            "but it's '%s'.", str(self.covmat))
            except IOError:
                raise LoggedError(self.log, "Can't open covmat file '%s'.", self.covmat)
            if header[0] != "#":
                raise LoggedError(
                    self.log, "The first line of the covmat file '%s' "
                              "must be one list of parameter names separated by spaces "
                              "and staring with '#', and the rest must be a square "
                              "matrix, with one row per line.", self.covmat)
            loaded_params = header.strip("#").strip().split()
        elif hasattr(self.covmat, "__getitem__"):
            if not self.covmat_params:
                raise LoggedError(
                    self.log, "If a covariance matrix is passed as a numpy array, "
                              "you also need to pass the parameters it corresponds to "
                              "via 'covmat_params: [name1, name2, ...]'.")
            loaded_params = self.covmat_params
            loaded_covmat = np.array(self.covmat)
        elif self.covmat:
            raise LoggedError(self.log, "Invalid covmat")
        if self.covmat is not None:
            str_msg = "the `covmat_params` list"
            if isinstance(self.covmat, str):
                str_msg = "the header of the covmat file %r" % self.covmat
            if len(loaded_params) != len(set(loaded_params)):
                duplicated = list(set(
                    p for p in loaded_params if list(loaded_params).count(p) > 1))
                raise LoggedError(
                    self.log,
                    "Parameter(s) %r appear more than once in %s", duplicated, str_msg)
            if len(loaded_params) != loaded_covmat.shape[0]:
                raise LoggedError(
                    self.log, "The number of parameters in %s and the "
                              "dimensions of the matrix do not agree: %d vs %r",
                    str_msg, len(loaded_params), loaded_covmat.shape)
            loaded_covmat = np.atleast_2d(loaded_covmat)
            is_square_symmetric = (len(loaded_covmat.shape) == 2 and
                                   loaded_covmat.shape[0] == loaded_covmat.shape[1] and
                                   np.allclose(loaded_covmat.T, loaded_covmat))
            # Not checking for positive-definiteness yet: may contain highly degenerate
            # derived parameters that would spoil it now, but will later be dropped.
            if not is_square_symmetric:
                from_msg = (f"loaded from '{self.covmat}'" if isinstance(self.covmat, str)
                            else "passed")
                raise LoggedError(
                    self.log,
                    f"The covariance matrix {from_msg} is not a symmetric square matrix.")
            # Fill with parameters in the loaded covmat
            renames = {p: [p] + str_to_list(v.get("renames") or [])
                       for p, v in params_infos.items()}
            indices_used, indices_sampler = zip(*[
                [loaded_params.index(p),
                 [list(params_infos).index(q) for q, a in renames.items() if p in a]]
                for p in loaded_params])
            if not any(indices_sampler):
                raise LoggedError(
                    self.log,
                    "A proposal covariance matrix has been loaded, but none of its "
                    "parameters are actually sampled here. Maybe a mismatch between"
                    " parameter names in the covariance matrix and the input file?")
            indices_used, indices_sampler = zip(*[
                [i, j] for i, j in zip(indices_used, indices_sampler) if j])
            if any(len(j) - 1 for j in indices_sampler):
                first = next(j for j in indices_sampler if len(j) > 1)
                raise LoggedError(
                    self.log,
                    "The parameters %s have duplicated aliases. Can't assign them an "
                    "element of the covariance matrix unambiguously.",
                    ", ".join([list(params_infos)[i] for i in first]))
            indices_sampler = tuple(chain(*indices_sampler))
            covmat[np.ix_(indices_sampler, indices_sampler)] = (
                loaded_covmat[np.ix_(indices_used, indices_used)])
            self.log.info(
                "Covariance matrix loaded for params %r",
                [list(params_infos)[i] for i in indices_sampler])
            missing_params = set(params_infos).difference(
                list(params_infos)[i] for i in indices_sampler)
            if missing_params:
                self.log.info(
                    "Missing proposal covariance for params %r",
                    [p for p in self.model.parameterization.sampled_params()
                     if p in missing_params])
            else:
                self.log.info("All parameters' covariance loaded from given covmat.")
        # Fill gaps with "proposal" property, if present, otherwise ref (or prior)
        where_nan = np.isnan(covmat.diagonal())
        if np.any(where_nan):
            covmat[where_nan, where_nan] = np.array(
                [(info.get("proposal", np.nan) or np.nan) ** 2
                 for info in params_infos.values()])[where_nan]
        where_nan2 = np.isnan(covmat.diagonal())
        if np.any(where_nan2):
            # the variances are likely too large for a good proposal, e.g. conditional
            # widths may be much smaller than the marginalized ones.
            # Divide by 4, better to be too small than too large.
            covmat[where_nan2, where_nan2] = (self.model.prior.reference_variances()
                                              [where_nan2] / self.fallback_covmat_scale)
        assert not np.any(np.isnan(covmat))
        return covmat, where_nan

    def covmat_filename(self):
        if self.output:
            return os.path.join(
                self.output.folder, self.output.prefix + Extension.covmat)
        return None

    def dump_covmat(self, covmat=None):
        if covmat is None:
            covmat = self.covmat
        np.savetxt(self.covmat_filename(), covmat, header=" ".join(
            list(self.model.parameterization.sampled_params())))



import numpy as np
from typing import Any
from copy import deepcopy
from cobaya.model import get_model
from cobaya.theory import Theory
from cobaya.likelihood import LikelihoodInterface, Likelihood
from cobaya.typing import InputDict, empty_dict, ParamDict
from .common import process_packages_path
from .conftest import install_test_wrapper
from .test_cosmo_camb import get_camb


# Test separating out the BBN consistency constraint into separate theory code,
# using CAMB's BBN interpolator class. Tests dependencies/multi-theory with one
# agnostic theory

class BBN(Theory):
    bbn: Any

    def get_requirements(self):
        return {'ombh2', 'nnu'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        yhe = self.bbn.Y_He(self.provider.get_param('ombh2'),
                            self.provider.get_param('nnu') - 3.044)
        state['derived'] = {'YHe': yhe}

    def get_can_provide_params(self):
        return ['YHe']


class BBN2(Theory):
    params = {'ombh2': None, 'nnu': None, 'YHe': {'derived': True}}
    bbn: Any

    def calculate(self, state, want_derived=True, **params_values_dict):
        if want_derived:
            state['derived'] = {'YHe': self.bbn.Y_He(params_values_dict['ombh2'],
                                                     params_values_dict['nnu'] - 3.044)}


# noinspection PyDefaultArgument
def cmb_likelihood(_self):
    results = _self.provider.get_CAMBdata()
    derived = {'check': results.Params.YHe}
    return results.Params.YHe, derived


cmb_likelihood_info = {'external': cmb_likelihood, 'output_params': 'check',
                       'requires': {'Hubble': {'z': [0.5]}, 'CAMBdata': None}}

camb_params = {
    "ombh2": 0.022274,
    "omch2": 0.11913,
    "cosmomc_theta": 0.01040867,
    "As": 0.2132755716e-8,
    "ns": 0.96597,
    "tau": 0.0639,
    "nnu": 3.044}

bbn_table = "PRIMAT_Yp_DH_Error.dat"
debug = True
info: InputDict = {'likelihood': {'cmb': cmb_likelihood_info},
                   'theory': {
                       'camb': {"extra_args": {"lens_potential_accuracy": 1},
                                "requires": ['YHe', 'ombh2']},
                       'bbn': {'external': BBN, 'provides': ['YHe']}},
                   'params': camb_params,
                   'debug': debug, 'stop_at_error': True}

info2: InputDict = {'likelihood': {'cmb': cmb_likelihood_info},
                    'theory': {'camb': {"requires": ['YHe', 'ombh2']}, 'bbn': BBN2},
                    'params': camb_params, 'debug': debug}


def test_bbn_yhe(packages_path, skip_not_installed):
    packages_path = process_packages_path(packages_path)
    install_test_wrapper(skip_not_installed, get_camb, packages_path)
    from camb.bbn import BBN_table_interpolator
    BBN.bbn = BBN_table_interpolator(bbn_table)
    BBN2.bbn = BBN.bbn

    info['params']['check'] = {'derived': True}

    for inf in (info, info2):
        inf['packages_path'] = packages_path
        for order in [1, -1]:
            for explicit_derived in [None, None, {'derived': True}]:
                print(inf, order, explicit_derived)
                model = get_model(inf)
                loglike, derived = model.loglikes({})
                vals = set([BBN.bbn.Y_He(camb_params['ombh2'])] + derived)
                assert len(vals) == 1, \
                    "wrong Yhe value: %s" % vals
                inf['params']["YHe"] = explicit_derived
            inf['params'].pop('YHe')
            inf['theory'] = {p: v for p, v in reversed(list(inf['theory'].items()))}


# Not inherit from BBN to derive likelihoods that account for the theory error

class BBN_likelihood(BBN2, LikelihoodInterface):
    """
    Sample YHe and just calculate a direct theory likelihood
    """
    params = dict.fromkeys(['ombh2', 'nnu', 'YHe'])

    def calculate(self, state, want_derived=True, **params_values_dict):
        ombh2 = params_values_dict['ombh2']
        delta_neff = params_values_dict['nnu'] - 3.044
        yhemean = self.bbn.Y_He(ombh2, delta_neff)
        error = self.bbn.get('sig(Yp^BBN)', ombh2, delta_neff)
        state['logp'] = -(params_values_dict['YHe'] - yhemean) ** 2 / (2 * error ** 2)

    def get_can_provide_params(self):
        return {}


class BBN_with_theory_errors(BBN, LikelihoodInterface):
    """
    The BBN theory prediction has an error. So we can derive a likelihood for
    sampling over a separate error parameter to account for it if we approximate
    the distribution as Gaussian.
    """
    params = {'BBN_delta': {'prior': {'min': 0, 'max': 1},
                            'ref': dict(dist='norm', loc=0, scale=1)}}

    def calculate(self, state, want_derived=True, **params_values_dict):
        ombh2, nnu = self.provider.get_param(['ombh2', 'nnu'])
        delta_neff = nnu - 3.044
        yhemean = self.bbn.Y_He(ombh2, delta_neff)
        # get theory error (neglecting difference between Y_He and Yp^BBN)
        error = self.bbn.get('sig(Yp^BBN)', ombh2, delta_neff)
        yhe = yhemean + error * params_values_dict['BBN_delta']
        if want_derived:
            state['derived'] = {'YHe': yhe}
        # take the error to be Gaussian
        state['logp'] = -params_values_dict['BBN_delta'] ** 2 / 2


info_error: InputDict = {'likelihood': dict([('cmb', cmb_likelihood_info),
                                             ('BBN', BBN_likelihood)]),
                         'theory': {'camb': {"requires": ['YHe', 'ombh2']}},
                         'params': dict(YHe={'prior': {'min': 0, 'max': 1}},
                                        **camb_params),
                         'debug': debug}

info_error2: InputDict = {'likelihood': {'cmb': cmb_likelihood_info,
                                         'BBN': {'external': BBN_with_theory_errors,
                                                 'provides': 'YHe'}},
                          'theory': {
                              'camb': {"requires": ['YHe', 'ombh2'], 'extra_args': {
                                  'bbn_predictor': 'PArthENoPE_880.2_standard.dat'}}},
                          'params': dict(BBN_delta={'prior': {'min': -5, 'max': 5}},
                                         **camb_params),
                          'debug': debug}


def test_bbn_likelihood(packages_path, skip_not_installed):
    packages_path = process_packages_path(packages_path)
    install_test_wrapper(skip_not_installed, get_camb, packages_path)
    from camb.bbn import BBN_table_interpolator
    BBN_likelihood.bbn = BBN_table_interpolator(bbn_table)
    model = get_model(info_error, packages_path=packages_path)
    assert np.allclose(model.loglikes({'YHe': 0.246})[0], [0.246, -0.8696], rtol=1e-3), \
        "Failed BBN likelihood with %s" % info_error

    # second case, BBN likelihood has to be calculated before CAMB
    BBN_with_theory_errors.bbn = BBN_likelihood.bbn
    model = get_model(info_error2, packages_path=packages_path)
    assert np.allclose(model.loglikes({'BBN_delta': 1.0})[0], [0.24594834, -0.5],
                       rtol=1e-4)


class ExamplePrimordialPk(Theory):

    def initialize(self):
        # need to provide valid results at wide k range, any that might be used
        self.ks = np.logspace(-5.5, 2, 1000)

    def calculate(self, state, want_derived=True, **params_values_dict):
        pivot_scalar = 0.05
        pk = (self.ks / pivot_scalar) ** (
                params_values_dict['testns'] - 1) * params_values_dict['testAs']
        state['primordial_scalar_pk'] = {'kmin': self.ks[0], 'kmax': self.ks[-1],
                                         'Pk': pk, 'log_regular': True}

    def get_primordial_scalar_pk(self):
        return self.current_state['primordial_scalar_pk']

    def get_can_support_params(self):
        return ['testAs', 'testns']


testAs = 1.8e-9
testns = 0.8


class Pklike(Likelihood):
    def logp(self, **params_values):
        results = self.provider.get_CAMBdata()
        print(results.Params.scalar_power(1.1),
              testAs * (1.1 / 0.05) ** (testns - 1))
        assert np.allclose(results.Params.scalar_power(1.1),
                           testAs * (1.1 / 0.05) ** (testns - 1), rtol=1e-3, atol=1e-20)
        return 0

    def get_requirements(self):
        return {'Cl': {'tt': 1000}, 'CAMBdata': None}


info_pk: InputDict = {'likelihood': {'cmb': Pklike},
                      'theory': {'camb': {"external_primordial_pk": True},
                                 'my_pk': ExamplePrimordialPk},
                      'params': {
                          "ombh2": 0.022274,
                          "omch2": 0.11913,
                          "cosmomc_theta": 0.01040867,
                          "tau": 0.0639,
                          "nnu": 3.044,
                          'testAs': {'prior': {'min': 1e-9, 'max': 1e-8}},
                          'testns': {'prior': {'min': 0.8, 'max': 1.2}}
                      },
                      'stop_at_error': True,
                      'debug': debug}


def test_primordial_pk(packages_path, skip_not_installed):
    packages_path = process_packages_path(packages_path)
    info_pk['packages_path'] = packages_path
    model = install_test_wrapper(skip_not_installed, get_model, info_pk)
    model.loglikes({'testAs': testAs, 'testns': testns})


class BinnedPk(Theory):
    # example splined power spectrum exp(-2 tau)P(k) based on bin values.
    # Can pass dense sampling to CAMB to reproduce any function, here bins are directly
    # cubic spline values use by CAMB.
    # Note need to  have wide k bounds or latest CAMB (which takes value beyond start
    # and end bins to be equal to the end bins).
    nbins: int = 20
    k_min_bin: float = np.log10(0.001)
    k_max_bin: float = np.log10(0.35)
    scale: float = 1e-9
    bin_par: ParamDict = {'prior': {'min': 0, 'max': 100}}

    def initialize(self):
        self.ks = np.logspace(self.k_min_bin, self.k_max_bin, self.nbins)

    def get_requirements(self):
        return {'tau'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        pk = np.zeros_like(self.ks)
        for b in range(self.nbins):
            pk[b] = params_values_dict['b%s' % (b + 1)]
        pk *= self.scale * np.exp(2 * self.provider.get_param('tau'))
        # should use log_regular: True for speed if the binning is log regular
        # here test the non-regular option for coverage
        state['primordial_scalar_pk'] = {'k': self.ks,
                                         'Pk': pk, 'log_regular': False}

    def get_primordial_scalar_pk(self):
        return self.current_state['primordial_scalar_pk']

    @classmethod
    def get_class_options(cls, input_options=empty_dict):
        # dynamically generate defaults for params based on nbins
        options = super().get_class_options().copy()
        nbins = input_options['nbins']
        bin_par = input_options.get('bin_par', cls.bin_par)
        params = {}
        for b in range(nbins):
            par = deepcopy(bin_par.copy())
            par['latex'] = 'b_%s' % (b + 1)
            params['b%s' % (b + 1)] = par
        options['params'] = params
        return options


def test_pk_binning(packages_path, skip_not_installed):
    # reproduce power law by sending in spline point values
    # has to be fine sampling to get to 1e-3 precision in test.
    nbins = 40
    tau = 0.05
    k_min_bin = -5.5
    k_max_bin = 2

    _info: InputDict = \
        {'packages_path': process_packages_path(packages_path),
         'likelihood': {'cmb': Pklike},
         'theory': {'camb': {"external_primordial_pk": True},
                    'my_pk': {"external": BinnedPk,
                              'nbins': nbins, 'k_min_bin': k_min_bin,
                              'k_max_bin': k_max_bin
                              }},
         'params': {
             "ombh2": 0.022274,
             "omch2": 0.11913,
             "cosmomc_theta": 0.01040867,
             "tau": tau,
             "nnu": 3.044
         },
         'stop_at_error': True,
         'debug': debug}
    scale = 1e-9
    ks = np.logspace(k_min_bin, k_max_bin, nbins)

    def pk_test(k):
        return testAs * (k / 0.05) ** (testns - 1) / scale * np.exp(-2 * tau)

    pars = {'b%s' % (b + 1): pk_test(ks[b]) for b in range(nbins)}
    model = install_test_wrapper(skip_not_installed, get_model, _info)
    model.loglikes(pars)



"""
Tests the facility for importing external likelihoods.

It tests all possible input methods: callable and string
(direct evaluation and ``import_module``).

In each case, it tests the correctness of the values generated, and of the updated info.

The test likelihood is a gaussian half-ring, combined with a gaussian in one of the tests.

For manual testing, and observing/plotting the density, pass ``manual=True`` to
``body of test``.
"""

# Local
from .common_external import info_string, info_callable, info_mixed, info_import
from .common_external import info_derived, body_of_test


def test_likelihood_external_string(tmpdir):
    body_of_test(info_string, "likelihood", tmpdir)


def test_likelihood_external_callable(tmpdir):
    body_of_test(info_callable, "likelihood", tmpdir)


def test_likelihood_external_mixed(tmpdir):
    body_of_test(info_mixed, "likelihood", tmpdir)


def test_likelihood_external_import(tmpdir):
    body_of_test(info_import, "likelihood", tmpdir)


def test_likelihood_external_derived(tmpdir):
    body_of_test(info_derived, "likelihood", tmpdir, derived=True)



"""
.. module:: BoltzmannBase

:Synopsis: Template for Cosmological theory codes.
           Mostly here to document how to compute and get observables.
:Author: Jesus Torrado

"""
import numpy as np
from scipy.interpolate import RectBivariateSpline
from typing import Mapping, Iterable

# Local
from cobaya.theory import Theory
from cobaya.tools import deepcopy_where_possible, combine_1d, combine_2d, check_2d
from cobaya.log import LoggedError, abstract, get_logger
from cobaya.conventions import Const
from cobaya.typing import empty_dict, InfoDict

H_units_conv_factor = {"1/Mpc": 1, "km/s/Mpc": Const.c_km_s}


class BoltzmannBase(Theory):
    _is_abstract = True
    renames: Mapping[str, str] = empty_dict
    extra_args: InfoDict
    _must_provide: dict
    path: str

    def initialize(self):

        # Dict of named tuples to collect requirements and computation methods
        self.collectors = {}
        # Additional input parameters to pass to CAMB, and attributes to set_ manually
        self.extra_args = deepcopy_where_possible(self.extra_args) or {}
        self._must_provide = {}

    def initialize_with_params(self):
        self.check_no_repeated_input_extra()

    def get_allow_agnostic(self):
        return True

    def translate_param(self, p):
        return self.renames.get(p, p)

    def get_param(self, p):
        translated = self.translate_param(p)
        for pool in ["params", "derived", "derived_extra"]:
            value = (self.current_state[pool] or {}).get(translated, None)
            if value is not None:
                return value

        raise LoggedError(self.log, "Parameter not known: '%s'", p)

    def _norm_vars_pairs(self, vars_pairs, name):
        # Empty list: default to *total matter*: CMB + Baryon + MassiveNu
        vars_pairs = vars_pairs or [2 * ["delta_tot"]]
        if not isinstance(vars_pairs, Iterable):
            raise LoggedError(self.log, "vars_pairs must be an iterable of pairs "
                                        "of variable names: got '%r' for %s",
                              vars_pairs, name)
        if isinstance(list(vars_pairs)[0], str):
            vars_pairs = [vars_pairs]
        pairs = set()
        for pair in vars_pairs:
            if len(list(pair)) != 2 or not all(isinstance(x, str) for x in pair):
                raise LoggedError(self.log,
                                  "Cannot understand vars_pairs '%r' for %s",
                                  vars_pairs, name)
            pairs.add(tuple(sorted(pair)))
        return pairs

    def _check_args(self, req, value, vals):
        for par in vals:
            if par not in value:
                raise LoggedError(self.log, "%s must specify %s in requirements",
                                  req, par)

    def must_provide(self, **requirements):
        r"""
        Specifies the quantities that this Boltzmann code is requested to compute.

        Typical requisites in Cosmology (as keywords, case-insensitive):

        - ``Cl={...}``: CMB lensed power spectra, as a dictionary ``{[spectrum]:
          l_max}``, where the possible spectra are combinations of ``"t"``, ``"e"``,
          ``"b"`` and ``"p"`` (lensing potential). Get with :func:`~BoltzmannBase.get_Cl`.
        - ``unlensed_Cl={...}``: CMB unlensed power spectra, as a dictionary
          ``{[spectrum]: l_max}``, where the possible spectra are combinations of
          ``"t"``, ``"e"``, ``"b"``. Get with :func:`~BoltzmannBase.get_unlensed_Cl`.
        - **[BETA: CAMB only; notation may change!]** ``source_Cl={...}``:
          :math:`C_\ell` of given sources with given windows, e.g.:
          ``source_name: {"function": "spline"|"gaussian", [source_args]``;
          for now, ``[source_args]`` follows the notation of `CAMBSources
          <https://camb.readthedocs.io/en/latest/sources.html>`_.
          It can also take ``"lmax": [int]``, ``"limber": True`` if Limber approximation
          desired, and ``"non_linear": True`` if non-linear contributions requested.
          Get with :func:`~BoltzmannBase.get_source_Cl`.
        - ``Pk_interpolator={...}``: Matter power spectrum interpolator in :math:`(z, k)`.
          Takes ``"z": [list_of_evaluated_redshifts]``, ``"k_max": [k_max]``,
          ``"nonlinear": [True|False]``,
          ``"vars_pairs": [["delta_tot", "delta_tot"], ["Weyl", "Weyl"], [...]]}``.
          Notice that ``k_min`` cannot be specified. To reach a lower one, use
          ``extra_args`` in CAMB to increase ``accuracyboost`` and ``TimeStepBoost``, or
          in CLASS to decrease ``k_min_tau0``. The method :func:`~get_Pk_interpolator` to
          retrieve the interpolator admits extrapolation limits ``extrap_[kmax|kmin]``.
          It is recommended to use ``extrap_kmin`` to reach the desired ``k_min``, and
          increase precision parameters only as much as necessary to improve over the
          interpolation. All :math:`k` values should be in units of
          :math:`1/\mathrm{Mpc}`.

          Non-linear contributions are included by default. Note that the non-linear
          setting determines whether non-linear corrections are calculated; the
          :func:`~get_Pk_interpolator` method also has a non-linear argument to specify if
          you want the linear or non-linear spectrum returned (to have both linear and
          non-linear spectra available request a tuple ``(False,True)`` for the non-linear
          argument).
        - ``Pk_grid={...}``: similar to ``Pk_interpolator`` except that rather than
          returning a bicubic spline object it returns the raw power spectrum grid as
          a ``(k, z, P(z,k))`` set of arrays. Get with :func:`~BoltzmannBase.get_Pk_grid`.
        - ``sigma_R={...}``: RMS linear fluctuation in spheres of radius :math:`R` at
          redshifts :math:`z`. Takes ``"z": [list_of_evaluated_redshifts]``,
          ``"k_max": [k_max]``, ``"vars_pairs": [["delta_tot", "delta_tot"],  [...]]``,
          ``"R": [list_of_evaluated_R]``. Note that :math:`R` is in :math:`\mathrm{Mpc}`,
          not :math:`h^{-1}\,\mathrm{Mpc}`. Get with :func:`~BoltzmannBase.get_sigma_R`.
        - ``Hubble={'z': [z_1, ...]}``: Hubble rates at the requested redshifts.
          Get it with :func:`~BoltzmannBase.get_Hubble`.
        - ``Omega_b={'z': [z_1, ...]}``: Baryon density parameter at the requested
          redshifts. Get it with :func:`~BoltzmannBase.get_Omega_b`.
        - ``Omega_cdm={'z': [z_1, ...]}``: Cold dark matter density parameter at the
          requested redshifts. Get it with :func:`~BoltzmannBase.get_Omega_cdm`.
        - ``Omega_nu_massive={'z': [z_1, ...]}``: Massive neutrinos' density parameter at
          the requested redshifts. Get it with
          :func:`~BoltzmannBase.get_Omega_nu_massive`.
        - ``angular_diameter_distance={'z': [z_1, ...]}``: Physical angular
          diameter distance to the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_angular_diameter_distance`.
        - ``angular_diameter_distance_2={'z_pairs': [(z_1a, z_1b), (z_2a, z_2b)...]}``:
          Physical angular diameter distance between the pairs of redshifts requested.
          If a 1d array of redshifts is passed as `z_pairs`, all possible combinations
          of two are computed and stored (not recommended if only a subset is needed).
          Get it with :func:`~BoltzmannBase.get_angular_diameter_distance_2`.
        - ``comoving_radial_distance={'z': [z_1, ...]}``: Comoving radial distance
          from us to the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_comoving_radial_distance`.
        - ``sigma8_z={'z': [z_1, ...]}``: Amplitude of rms fluctuations
          :math:`\sigma_8` at the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_sigma8`.
        - ``fsigma8={'z': [z_1, ...]}``: Structure growth rate
          :math:`f\sigma_8` at the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_fsigma8`.
        - **Other derived parameters** that are not included in the input but whose
          value the likelihood may need.

        """
        super().must_provide(**requirements)
        self._must_provide = self._must_provide or dict.fromkeys(self.output_params or [])
        # Accumulate the requirements across several calls in a safe way;
        # e.g. take maximum of all values of a requested precision parameter
        for k, v in requirements.items():
            # Products and other computations
            if k == "Cl":
                current = self._must_provide.get(k, {})
                v = {cl.lower(): v[cl] for cl in v}  # to lowercase
                self._must_provide[k] = {
                    cl.lower(): max(current.get(cl.lower(), 0), v.get(cl, 0))
                    for cl in set(current).union(v)}
            elif k == "unlensed_Cl":
                current = self._must_provide.get(k, {})
                self._must_provide[k] = {
                    cl.lower(): max(current.get(cl.lower(), 0), v.get(cl, 0))
                    for cl in set(current).union(v)}
            elif k == 'sigma_R':
                self._check_args(k, v, ('z', 'R'))
                for pair in self._norm_vars_pairs(v.pop("vars_pairs", []), k):
                    k = ("sigma_R",) + pair
                    current = self._must_provide.get(k, {})
                    self._must_provide[k] = {
                        "R": combine_1d(v["R"], current.get("R")),
                        "z": combine_1d(v["z"], current.get("z")),
                        "k_max": max(current.get("k_max", 0),
                                     v.get("k_max", 2 / np.min(v["R"])))}
            elif k in ("Pk_interpolator", "Pk_grid"):
                # arguments are all identical, collect all in Pk_grid
                self._check_args(k, v, ('z', 'k_max'))
                redshifts = v.pop("z")
                k_max = v.pop("k_max")
                nonlin = v.pop("nonlinear", True)
                if not isinstance(nonlin, Iterable):
                    nonlin = [nonlin]
                for var_pair in self._norm_vars_pairs(v.pop("vars_pairs", []), k):
                    for nonlinear in nonlin:
                        k = ("Pk_grid", bool(nonlinear)) + var_pair
                        current = self._must_provide.get(k, {})
                        self._must_provide[k] = dict(
                            nonlinear=nonlinear,
                            z=combine_1d(redshifts, current.get("z")),
                            k_max=max(current.get("k_max", 0), k_max), **v)
            elif k == "source_Cl":
                if k not in self._must_provide:
                    self._must_provide[k] = {}
                if "sources" not in v:
                    raise LoggedError(
                        self.log, "Needs a 'sources' key, containing a dict with every "
                                  "source name and definition")
                # Check that no two sources with equal name but diff specification
                # for source, window in v["sources"].items():
                #     if source in (getattr(self, "sources", {}) or {}):
                #         # TODO: improve this test!!!
                #         # (2 z-vectors that fulfill np.allclose would fail a == test)
                #         if window != self.sources[source]:
                #             raise LoggedError(
                #                 self.log,
                #                 "Source %r requested twice with different specification"
                #                 ": %r vs %r.", source, window, self.sources[source])
                self._must_provide[k].update(v)
            elif k in {"Hubble", "Omega_b", "Omega_cdm", "Omega_nu_massive",
                       "angular_diameter_distance", "comoving_radial_distance",
                       "sigma8_z", "fsigma8"}:
                if k not in self._must_provide:
                    self._must_provide[k] = {}
                if not isinstance(v, Iterable) or "z" not in v:
                    raise LoggedError(
                        self.log,
                        f"The value in the dictionary of requisites {k} must be a "
                        "dictionary containing the key 'z' with a list of redshifts "
                        f"(got instead {{{k}: {v}}})"
                    )
                self._must_provide[k]["z"] = combine_1d(
                    v["z"], self._must_provide[k].get("z"))
            elif k == "angular_diameter_distance_2":
                if k not in self._must_provide:
                    self._must_provide[k] = {}
                zs = v.pop("z_pairs", None)
                # Combine pairs with previous ones and uniquify
                try:
                    self._must_provide[k]["z_pairs"] = combine_2d(
                        zs, self._must_provide[k].get("z_pairs"))
                except ValueError:
                    raise LoggedError(
                        self.log, "For requisite `angular_diameter_distance2`, `z_pairs` "
                                  "must be a list of pairs (z1, z2), but is "
                                  f"{v['z_pairs']}")
            # Extra derived parameters and other unknown stuff (keep capitalization)
            elif v is None:
                self._must_provide[k] = None
            else:
                raise LoggedError(self.log, "Unknown required product: '%s'.", k)

    def requested(self):
        """
        Returns the full set of requested cosmological products and parameters.
        """
        return self._must_provide

    def check_no_repeated_input_extra(self):
        """
        Checks that there are no repeated parameters between input and extra.

        Should be called at initialisation, and at the end of every call to must_provide()
        """
        common = set(self.input_params).intersection(self.extra_args)
        if common:
            raise LoggedError(
                self.log, "The following parameters appear both as input parameters and "
                          "as extra arguments: %s. Please, remove one of the definitions "
                          "of each.", common)

    def _get_z_dependent(self, quantity, z, pool=None):
        if pool is None:
            pool = self.collectors[quantity].z_pool
        try:
            i_kwarg_z = pool.find_indices(z)
        except ValueError:
            raise LoggedError(self.log, f"{quantity} not computed for all z requested. "
                                        f"Requested z are {z}, but computed ones are "
                                        f"{pool.values}.")
        return np.array(self.current_state[quantity], copy=True)[i_kwarg_z]

    def _get_z_pair_dependent(self, quantity, z_pairs, inv_value=0):
        """
        ``inv_value`` (default=0) is assigned to pairs for which ``z1 > z2``.
        """
        try:
            check_2d(z_pairs, allow_1d=False)
        except ValueError:
            raise LoggedError(self.log, f"z_pairs={z_pairs} not correctly formatted for "
                                        f"{quantity}. It should be a list of pairs.")
        # Only recover for correctly sorted pairs
        z_pairs_arr = np.array(z_pairs)
        i_right = (z_pairs_arr[:, 0] <= z_pairs_arr[:, 1])
        pool = self.collectors[quantity].z_pool
        try:
            i_z_pair = pool.find_indices(z_pairs_arr[i_right])
        except ValueError:
            raise LoggedError(
                self.log, f"{quantity} not computed for all z pairs requested. "
                          f"Requested z are {z_pairs}, but computed ones are "
                          f"{pool.values}.")
        result = np.full(len(z_pairs), inv_value, dtype=float)
        result[i_right] = np.array(self.current_state[quantity], copy=True)[i_z_pair]
        return result

    def _cmb_unit_factor(self, units, T_cmb):
        units_factors = {"1": 1,
                         "muK2": T_cmb * 1.e6,
                         "K2": T_cmb,
                         "FIRASmuK2": 2.7255e6,
                         "FIRASK2": 2.7255
                         }
        try:
            return units_factors[units]
        except KeyError:
            raise LoggedError(self.log, "Units '%s' not recognized. Use one of %s.",
                              units, list(units_factors))

    @abstract
    def get_Cl(self, ell_factor=False, units="FIRASmuK2"):
        r"""
        Returns a dictionary of lensed CMB power spectra and the lensing potential ``pp``
        power spectrum.

        Set the units with the keyword ``units=number|'muK2'|'K2'|'FIRASmuK2'|'FIRASK2'``.
        The default is ``FIRASmuK2``, which returns CMB :math:`C_\ell`'s in
        FIRAS-calibrated :math:`\mu K^2`, i.e. scaled by a fixed factor of
        :math:`(2.7255\cdot 10^6)^2` (except for the lensing potential power spectrum,
        which is always unitless).
        The ``muK2`` and ``K2`` options use the model's CMB temperature.

        If ``ell_factor=True`` (default: ``False``), multiplies the spectra by
        :math:`\ell(\ell+1)/(2\pi)` (or by :math:`\ell^2(\ell+1)^2/(2\pi)` in the case of
        the lensing potential ``pp`` spectrum).
        """

    @abstract
    def get_unlensed_Cl(self, ell_factor=False, units="FIRASmuK2"):
        r"""
        Returns a dictionary of unlensed CMB power spectra.

        For ``units`` options, see :func:`~BoltzmannBase.get_Cl`.

        If ``ell_factor=True`` (default: ``False``), multiplies the spectra by
        :math:`\ell(\ell+1)/(2\pi)`.
        """

    def get_Hubble(self, z, units="km/s/Mpc"):
        r"""
        Returns the Hubble rate at the given redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        The available units are ``"km/s/Mpc"`` (i.e. :math:`cH(\mathrm(Mpc)^{-1})`) and
        ``1/Mpc``.
        """
        try:
            return self._get_z_dependent("Hubble", z) * H_units_conv_factor[units]
        except KeyError:
            raise LoggedError(
                self.log, "Units not known for H: '%s'. Try instead one of %r.",
                units, list(H_units_conv_factor))

    def get_Omega_b(self, z):
        r"""
        Returns the Baryon density parameter at the given redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("Omega_b", z)

    def get_Omega_cdm(self, z):
        r"""
        Returns the Cold Dark Matter density parameter at the given redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("Omega_cdm", z)

    def get_Omega_nu_massive(self, z):
        r"""
        Returns the Massive neutrinos' density parameter at the given redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("Omega_nu_massive", z)

    def get_angular_diameter_distance(self, z):
        r"""
        Returns the physical angular diameter distance in :math:`\mathrm{Mpc}` to the
        given redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("angular_diameter_distance", z)

    def get_angular_diameter_distance_2(self, z_pairs):
        r"""
        Returns the physical angular diameter distance between pairs of redshifts
        `z_pairs` in :math:`\mathrm{Mpc}`.

        The redshift pairs must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.

        Return zero for pairs in which ``z1 > z2``.
        """
        return self._get_z_pair_dependent(
            "angular_diameter_distance_2", z_pairs, inv_value=0)

    def get_comoving_radial_distance(self, z):
        r"""
        Returns the comoving radial distance in :math:`\mathrm{Mpc}` to the given
        redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("comoving_radial_distance", z)

    def get_Pk_interpolator(self, var_pair=("delta_tot", "delta_tot"), nonlinear=True,
                            extrap_kmin=None, extrap_kmax=None):
        r"""
        Get a :math:`P(z,k)` bicubic interpolation object
        (:class:`PowerSpectrumInterpolator`).

        In the interpolator returned, the input :math:`k` and resulting
        :math:`P(z,k)` are in units of :math:`1/\mathrm{Mpc}` and
        :math:`\mathrm{Mpc}^3` respectively (not in :math:`h^{-1}` units).

        :param var_pair: variable pair for power spectrum
        :param nonlinear: non-linear spectrum (default True)
        :param extrap_kmin: use log linear extrapolation from ``extrap_kmin`` up to min
                            :math:`k`.
        :param extrap_kmax: use log linear extrapolation beyond max :math:`k` computed up
                            to ``extrap_kmax``.
        :return: :class:`PowerSpectrumInterpolator` instance.
        """
        nonlinear = bool(nonlinear)
        key = (("Pk_interpolator", nonlinear, extrap_kmin, extrap_kmax) +
               tuple(sorted(var_pair)))
        if key in self.current_state:
            return self.current_state[key]
        k, z, pk = self.get_Pk_grid(var_pair=var_pair, nonlinear=nonlinear)
        log_p = True
        sign = 1
        if np.any(pk < 0):
            if np.all(pk < 0):
                sign = -1
            else:
                log_p = False
        extrapolating = ((extrap_kmax and extrap_kmax > k[-1]) or
                         (extrap_kmin and extrap_kmin < k[0]))
        if log_p:
            pk = np.log(sign * pk)
        elif extrapolating:
            raise LoggedError(self.log,
                              'Cannot do log extrapolation with zero-crossing pk '
                              'for %s, %s' % var_pair)
        result = PowerSpectrumInterpolator(
            z, k, pk, logP=log_p, logsign=sign,
            extrap_kmin=extrap_kmin, extrap_kmax=extrap_kmax)
        self.current_state[key] = result
        return result

    def get_Pk_grid(self, var_pair=("delta_tot", "delta_tot"), nonlinear=True):
        r"""
        Get  matter power spectrum, e.g. suitable for splining.
        Returned arrays may be bigger or more densely sampled than requested, but will
        include required values.

        In the grid returned, :math:`k` and :math:`P(z,k)` are in units of
        :math:`1/\mathrm{Mpc}` and :math:`\mathrm{Mpc}^3` respectively
        (not in :math:`h^{-1}` units), and :math:`z` and :math:`k` are in
        **ascending** order.

        :param nonlinear: whether the linear or non-linear spectrum
        :param var_pair: which power spectrum
        :return: ``k``, ``z``, ``Pk``, where ``k`` and ``z`` are 1-d arrays,
                 and the 2-d array ``Pk[i,j]`` is the value of :math:`P(z,k)` at ``z[i]``,
                 ``k[j]``.
        """
        try:
            return self.current_state[
                ("Pk_grid", bool(nonlinear)) + tuple(sorted(var_pair))]
        except KeyError:
            if ("Pk_grid", False) + tuple(sorted(var_pair)) in self.current_state:
                raise LoggedError(self.log,
                                  "Getting non-linear matter power but 'nonlinear' "
                                  "not specified in requirements")
            raise LoggedError(self.log, "Matter power %s, %s not computed" % var_pair)

    def get_sigma_R(self, var_pair=("delta_tot", "delta_tot")):
        r"""
        Get :math:`\sigma_R(z)`, the RMS power in a sphere of radius :math:`R` at
        redshift :math:`z`.

        Note that :math:`R` is in :math:`\mathrm{Mpc}`, not :math:`h^{-1}\,\mathrm{Mpc}`,
        and :math:`z` and :math:`R` are returned in **ascending** order.

        You may get back more values than originally requested, but the requested
        :math:`R` and :math:`z` should be in the returned arrays.

        :param var_pair: which two fields to use for the RMS power
        :return: ``z``, ``R``, ``sigma_R``, where ``z`` and ``R`` are arrays of computed
                 values, and ``sigma_R[i,j]`` is the value :math:`\sigma_R(z)` for
                 ``z[i]``, ``R[j]``.
        """
        try:
            return self.current_state[("sigma_R",) + tuple(sorted(var_pair))]
        except KeyError:
            raise LoggedError(self.log, "sigmaR %s not computed" % var_pair)

    @abstract
    def get_source_Cl(self):
        r"""
        Returns a dict of power spectra of for the computed sources, with keys a tuple of
        sources ``([source1], [source2])``, and an additional key ``ell`` containing the
        multipoles.
        """

    def get_sigma8_z(self, z):
        r"""
        Present day linear theory root-mean-square amplitude of the matter
        fluctuation spectrum averaged in spheres of radius 8 h^{−1} Mpc.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("sigma8_z", z)

    def get_fsigma8(self, z):
        r"""
        Structure growth rate :math:`f\sigma_8`, as defined in eq. 33 of
        `Planck 2015 results. XIII.
        Cosmological parameters <https://arxiv.org/pdf/1502.01589.pdf>`_,
        at the given redshift(s) ``z``.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("fsigma8", z)

    def get_auto_covmat(self, params_info, likes_info, random_state=None):
        r"""
        Tries to get match to a database of existing covariance matrix files for the
        current model and data.

        ``params_info`` should contain preferably the slow parameters only.
        """
        from cobaya.cosmo_input import get_best_covmat_ext
        return get_best_covmat_ext(self.packages_path, params_info, likes_info,
                                   random_state)


class PowerSpectrumInterpolator(RectBivariateSpline):
    r"""
    2D spline interpolation object (scipy.interpolate.RectBivariateSpline)
    to evaluate matter power spectrum as function of z and k.

    *This class is adapted from CAMB's own P(k) interpolator, by Antony Lewis;
    it's mostly interface-compatible with the original.*

    :param z: values of z for which the power spectrum was evaluated.
    :param k: values of k for which the power spectrum was evaluated.
    :param P_or_logP: Values of the power spectrum (or log-values, if logP=True).
    :param logP: if True (default: False), log of power spectrum are given and used
        for the underlying interpolator.
    :param logsign: if logP is True, P_or_logP is log(logsign*Pk)
    :param extrap_kmax: if set, use power law extrapolation beyond kmax up to
        extrap_kmax; useful for tails of integrals.
    """

    def __init__(self, z, k, P_or_logP, extrap_kmin=None, extrap_kmax=None, logP=False,
                 logsign=1):
        self.islog = logP
        #  Check order
        z, k = (np.atleast_1d(x) for x in [z, k])
        if len(z) < 4:
            raise ValueError('Require at least four redshifts for Pk interpolation.'
                             'Consider using Pk_grid if you just need a small number'
                             'of specific redshifts (doing 1D splines in k yourself).')
        z, k, P_or_logP = np.array(z), np.array(k), np.array(P_or_logP)
        i_z = np.argsort(z)
        i_k = np.argsort(k)
        self.logsign = logsign
        self.z, self.k, P_or_logP = z[i_z], k[i_k], P_or_logP[i_z, :][:, i_k]
        self.zmin, self.zmax = self.z[0], self.z[-1]
        self.extrap_kmin, self.extrap_kmax = extrap_kmin, extrap_kmax
        logk = np.log(self.k)
        # Start from extrap_kmin using a (log,log)-linear extrapolation
        if extrap_kmin and extrap_kmin < self.input_kmin:
            if not logP:
                raise ValueError('extrap_kmin must use logP')
            logk = np.hstack(
                [np.log(extrap_kmin),
                 np.log(self.input_kmin) * 0.1 + np.log(extrap_kmin) * 0.9, logk])
            logPnew = np.empty((P_or_logP.shape[0], P_or_logP.shape[1] + 2))
            logPnew[:, 2:] = P_or_logP
            diff = (logPnew[:, 3] - logPnew[:, 2]) / (logk[3] - logk[2])
            delta = diff * (logk[2] - logk[0])
            logPnew[:, 0] = logPnew[:, 2] - delta
            logPnew[:, 1] = logPnew[:, 2] - delta * 0.9
            P_or_logP = logPnew
        # Continue until extrap_kmax using a (log,log)-linear extrapolation
        if extrap_kmax and extrap_kmax > self.input_kmax:
            if not logP:
                raise ValueError('extrap_kmax must use logP')
            logk = np.hstack(
                [logk, np.log(self.input_kmax) * 0.1 + np.log(extrap_kmax) * 0.9,
                 np.log(extrap_kmax)])
            logPnew = np.empty((P_or_logP.shape[0], P_or_logP.shape[1] + 2))
            logPnew[:, :-2] = P_or_logP
            diff = (logPnew[:, -3] - logPnew[:, -4]) / (logk[-3] - logk[-4])
            delta = diff * (logk[-1] - logk[-3])
            logPnew[:, -1] = logPnew[:, -3] + delta
            logPnew[:, -2] = logPnew[:, -3] + delta * 0.9
            P_or_logP = logPnew
        super().__init__(self.z, logk, P_or_logP)

    @property
    def input_kmin(self):
        """Minimum k for the interpolation (not incl. extrapolation range)."""
        return self.k[0]

    @property
    def input_kmax(self):
        """Maximum k for the interpolation (not incl. extrapolation range)."""
        return self.k[-1]

    @property
    def kmin(self):
        """Minimum k of the interpolator (incl. extrapolation range)."""
        if self.extrap_kmin is None:
            return self.input_kmin
        return self.extrap_kmin

    @property
    def kmax(self):
        """Maximum k of the interpolator (incl. extrapolation range)."""
        if self.extrap_kmax is None:
            return self.input_kmax
        return self.extrap_kmax

    def check_ranges(self, z, k):
        """Checks that we are not trying to extrapolate beyond the interpolator limits."""
        z = np.atleast_1d(z).flatten()
        min_z, max_z = min(z), max(z)
        if min_z < self.zmin and not np.allclose(min_z, self.zmin):
            raise LoggedError(get_logger(self.__class__.__name__),
                              f"Not possible to extrapolate to z={min(z)} "
                              f"(minimum z computed is {self.zmin}).")
        if max_z > self.zmax and not np.allclose(max_z, self.zmax):
            raise LoggedError(get_logger(self.__class__.__name__),
                              f"Not possible to extrapolate to z={max(z)} "
                              f"(maximum z computed is {self.zmax}).")
        k = np.atleast_1d(k).flatten()
        min_k, max_k = min(k), max(k)
        if min_k < self.kmin and not np.allclose(min_k, self.kmin):
            raise LoggedError(get_logger(self.__class__.__name__),
                              f"Not possible to extrapolate to k={min(k)} 1/Mpc "
                              f"(minimum k possible is {self.kmin} 1/Mpc).")
        if max_k > self.kmax and not np.allclose(max_k, self.kmax):
            raise LoggedError(get_logger(self.__class__.__name__),
                              f"Not possible to extrapolate to k={max(k)} 1/Mpc "
                              f"(maximum k possible is {self.kmax} 1/Mpc).")

    def P(self, z, k, grid=None):
        """
        Get the power spectrum at (z,k).
        """
        self.check_ranges(z, k)
        if grid is None:
            grid = not np.isscalar(z) and not np.isscalar(k)
        if self.islog:
            return self.logsign * np.exp(self(z, np.log(k), grid=grid, warn=False))
        else:
            return self(z, np.log(k), grid=grid, warn=False)

    def logP(self, z, k, grid=None):
        """
        Get the log power spectrum at (z,k). (or minus log power spectrum if
        islog and logsign=-1)
        """
        self.check_ranges(z, k)
        if grid is None:
            grid = not np.isscalar(z) and not np.isscalar(k)
        if self.islog:
            return self(z, np.log(k), grid=grid, warn=False)
        else:
            return np.log(self(z, np.log(k), grid=grid, warn=False))

    def __call__(self, *args, warn=True, **kwargs):
        if warn:
            get_logger(self.__class__.__name__).warning(
                "Do not call the instance directly. Use instead methods P(z, k) or "
                "logP(z, k) to get the (log)power spectrum. (If you know what you are "
                "doing, pass warn=False)")
        return super().__call__(*args, **kwargs)



"""
.. module:: conventions

:Synopsis: Some physical constants and naming conventions
           to make the life of the maintainer easier.
:Author: Jesus Torrado

"""
from typing import Final

# Package name (for importlib)
# (apparently __package__ is only defined if you import something locally.
cobaya_package = __name__.rpartition('.')[0]


def get_version():
    from cobaya import __version__
    return __version__


debug_default = False
resume_default = False

kinds: Final = ("sampler", "theory", "likelihood")

# Reserved attributes for component classes with defaults.
# These are ignored by HasDefaults.get_class_options()
reserved_attributes: Final = {"input_params", "output_params", "install_options",
                              "bibtex_file", "file_base_name"}

# Conventional order for yaml dumping (purely cosmetic)
dump_sort_cosmetic: Final = ["theory", "likelihood", "prior", "params", "sampler", "post"]

# Separator for
# fields in parameter names and files
# Its manual inclusion in a string anywhere else (e.g. a parameter name) should be avoided
derived_par_name_separator: Final = "__"
separator_files: Final = "."


# Names for the samples' fields internally and in the output
class OutPar:
    weight = "weight"  # sample weight
    # minus log-posterior, or in general the total minus log-probability
    minuslogpost = "minuslogpost"
    minuslogprior = "minuslogprior"  # minus log-prior
    chi2 = "chi2"  # chi^2 = -2 * loglike (not always normalized to be useful)


prior_1d_name = "0"


def get_chi2_name(p):
    return OutPar.chi2 + derived_par_name_separator + str(p)


def undo_chi2_name(p):
    return p[len(OutPar.chi2 + derived_par_name_separator):]


def get_chi2_label(p):
    return r"\chi^2_\mathrm{" + str(p).replace("_", r"\ ") + "}"


def get_minuslogpior_name(piname):
    return OutPar.minuslogprior + derived_par_name_separator + piname


def get_minuslogprior_label(p):
    return r"-\log\pi_\mathrm{" + str(p).replace("_", r"\ ") + "}"


def minuslogprior_names(prior):
    return [get_minuslogpior_name(piname) for piname in prior]


def minuslogprior_labels(prior):
    return {piname: get_minuslogprior_label(piname) for piname in prior}


def chi2_names(likes):
    return [get_chi2_name(likename) for likename in likes]


def chi2_labels(likes):
    return {likename: get_chi2_label(likename) for likename in likes}


def minuslogpost_label():
    return r"-\log p"


# Output files

class FileSuffix:
    input = "input"
    updated = "updated"


class Extension:
    yaml = ".yaml"
    yamls = ".yaml", ".yml"
    dill = ".dill_pickle"
    checkpoint = ".checkpoint"
    progress = ".progress"
    covmat = ".covmat"
    evidence = ".logZ"


# Installation and container definitions
packages_path_arg: Final = "packages_path"
packages_path_input: Final = "packages_path"
packages_path_arg_posix: Final = packages_path_arg.replace("_", "-")
packages_path_env: Final = "COBAYA_PACKAGES_PATH"
packages_path_config_file = "config.yaml"
packages_path_containers = "/cobaya_packages"

install_skip_env: Final = "COBAYA_INSTALL_SKIP"
test_skip_env: Final = "COBAYA_TEST_SKIP"

products_path = "/products"

data_path = "data"
code_path = "code"

# Internal package structure
subfolders: Final = {"likelihood": "likelihoods",
                     "sampler": "samplers",
                     "theory": "theories"}

# Approximate overhead of cobaya per posterior evaluation. Useful for blocking speeds
overhead_time = 0.0003

# Line width for console printing
line_width = 120


# Physical constants (all definitions)
# ------------------
# Light speed
class Const:
    c_km_s = 299792.458  # speed of light
    h_J_s = 6.62607015e-34  # Planck's constant
    kB_J_K = 1.380649e-23  # Boltzmann constant



import sys
from cobaya.likelihood import Likelihood
from cobaya.theory import Theory
from cobaya.run import run
from cobaya.model import get_model
from cobaya.typing import InputDict, PostDict
from cobaya.log import LoggedError
from cobaya.post import post


if sys.version_info < (3, 8):
    print('Cobaya requires Python 3.8+, please upgrade.')
    sys.exit(1)

__author__ = "Jesus Torrado and Antony Lewis"
__version__ = "3.3"
__obsolete__ = False
__year__ = "2023"
__url__ = "https://cobaya.readthedocs.io"



r"""
.. module:: InstallableLikelihood

:Synopsis: Prototype class with class methods for simple installation of likelihood data.
:Author: Jesus Torrado and Antony Lewis

"""

# Global
import os
from packaging import version

# Local
from cobaya.likelihood import Likelihood
from cobaya.typing import InfoDict
from cobaya.log import get_logger
from cobaya.install import _version_filename
from cobaya.component import ComponentNotInstalledError
from cobaya.tools import VersionCheckError, resolve_packages_path


class InstallableLikelihood(Likelihood):
    """
    Prototype class for likelihood with installation and installation-check methods.
    """

    install_options: InfoDict = {}

    def __init__(self, *args, **kwargs):
        # Ensure check for install and version errors
        # (e.g. may inherit from a class that inherits from this one, and not have them)
        if self.install_options:
            name = self.get_qualified_class_name()
            logger = get_logger(name)
            packages_path = kwargs.get("packages_path") or resolve_packages_path()
            old = False
            try:
                installed = self.is_installed(path=packages_path)
            except Exception as excpt:  # catches VersionCheckError and unexpected ones
                installed = False
                old = isinstance(excpt, VersionCheckError)
                logger.error(f"{type(excpt).__name__}: {excpt}")
            if not installed:
                not_or_old = ("is not up to date" if old
                              else "has not been correctly installed")
                raise ComponentNotInstalledError(
                    logger, (f"The data for this likelihood {not_or_old}. To install it, "
                             f"run `cobaya-install {name}{' --upgrade' if old else ''}`"))
        super().__init__(*args, **kwargs)

    @classmethod
    def get_install_options(cls):
        """
        Returns class variables containing e.g. a download url, a version tag...
        """
        return cls.install_options

    @classmethod
    def get_path(cls, path):
        """
        Returns the (real) path where the package will be installed,
        given a global installation ``path``.

        Repeated recursive calls on the same global path produce the same output,
        i.e. ``cls.get_path(cls.get_path(path)) = cls.get_path(path)``.
        """
        opts = cls.get_install_options()
        if repo := opts.get("directory", opts.get("github_repository")):
            data_path = repo.split('/')[-1]
        else:
            data_path = opts.get("data_path", cls.__name__)
        install_path = os.path.realpath(os.path.join(path, "data", data_path))
        # Idempotent: return if input == proposed output
        if path.rstrip(os.sep).endswith(data_path.rstrip(os.sep)):
            return path
        return install_path

    @classmethod
    def is_installed(cls, **kwargs):
        """
        Performs an installation check and returns ``True`` if successful, ``False`` if
        not, or raises :class:`tools.VersionCheckError` if there is an obsolete
        installation.
        """
        if kwargs.get("data", True):
            path = cls.get_path(kwargs["path"])  # ensure full install path passed
            if not (opts := cls.get_install_options()):
                return True
            elif not os.path.exists(path) or not len(os.listdir(path)):
                log = get_logger(cls.get_qualified_class_name())
                (log.error if kwargs.get("show_error") is not False else log.info)(
                    "The given installation path does not exist: '%s'", path)
                return False
            elif release := opts.get("github_release"):
                try:
                    with open(os.path.join(path, _version_filename), "r") as f:
                        ver = f.readlines()[0]
                except FileNotFoundError:  # old install: no version file
                    ver = '0.0'
                if ((installed_version := version.parse(ver)) <
                        (min_version := version.parse(release))):
                    raise VersionCheckError(
                        f"Installed version ({installed_version}) "
                        f"older than minimum required one ({min_version}).")
        return True

    @classmethod
    def install(cls, path=None, data=True, no_progress_bars=False, **_kwargs):
        """
        Installs the necessary data packages for this likelihood.
        """
        if not data:
            return True
        log = get_logger(cls.get_qualified_class_name())
        if not (opts := cls.get_install_options()):
            log.info("No install options. Nothing to do.")
            return True
        if repo := opts.get("github_repository"):
            from cobaya.install import download_github_release
            log.info("Downloading %s data..." % repo)
            return download_github_release(
                os.path.join(path, "data"), repo, opts.get("github_release"),
                asset=opts.get("asset"), directory=opts.get("directory"),
                no_progress_bars=no_progress_bars, logger=log)
        else:
            full_path = cls.get_path(path)
            if not os.path.exists(full_path):
                os.makedirs(full_path)
            if not data:
                return True
            url = opts["download_url"]
            log.info("Downloading likelihood data file: %s...", url)
            from cobaya.install import download_file
            if not download_file(url, full_path, logger=log,
                                 no_progress_bars=no_progress_bars):
                return False
            log.info("Likelihood data downloaded and uncompressed correctly.")
            return True



# Tests to check correct input parsing and inheritance of defaults

# Global
from copy import deepcopy
import pytest
import os

# Local
from cobaya.typing import InputDict
from cobaya.run import run, run_script
from cobaya.log import LoggedError
from cobaya.input import get_default_info
from cobaya.yaml import yaml_dump_file, yaml_load_file

# Aux definitions and functions

test_info_common: InputDict = {
    "likelihood": {"_test": None},
    "sampler": {"evaluate": None}}


def test_prior_inherit_nonegiven():
    updated_info, _ = run(test_info_common)
    likname = list(test_info_common["likelihood"])[0]
    default_info = get_default_info(likname, "likelihood")
    assert updated_info["prior"] == default_info["prior"]


def test_prior_inherit_differentgiven():
    test_info = deepcopy(test_info_common)
    test_info["prior"] = {"third": "lambda a1: 1"}
    updated_info, _ = run(test_info)
    likname = list(test_info_common["likelihood"])[0]
    default_info = get_default_info(likname, "likelihood")
    default_info["prior"].update(test_info["prior"])
    assert updated_info["prior"] == default_info["prior"]


def test_prior_inherit_samegiven():
    test_info = deepcopy(test_info_common)
    likname = list(test_info_common["likelihood"])[0]
    default_info = get_default_info(likname, "likelihood")
    name, prior = deepcopy(default_info["prior"]).popitem()
    test_info["prior"] = {name: prior}
    updated_info, _ = run(test_info)
    assert updated_info["prior"] == default_info["prior"]


def test_prior_inherit_samegiven_differentdefinition():
    test_info = deepcopy(test_info_common)
    likname = list(test_info_common["likelihood"])[0]
    default_info = get_default_info(likname, "likelihood")
    name, prior = deepcopy(default_info["prior"]).popitem()
    test_info["prior"] = {name: "this is not a prior"}
    with pytest.raises(LoggedError):
        run(test_info)


def test_inherit_label_and_bounds():
    test_info = deepcopy(test_info_common)
    likname = list(test_info_common["likelihood"])[0]
    default_info_params = get_default_info(likname, "likelihood")["params"]
    test_info["params"] = deepcopy(default_info_params)
    test_info["params"]["a1"].pop("latex", None)
    # Remove limits, so they are inherited
    test_info = deepcopy(test_info_common)
    test_info["params"] = deepcopy(default_info_params)
    test_info["params"]["b1"].pop("min")
    test_info["params"]["b1"].pop("max")
    updated_info, _ = run(test_info)
    assert updated_info["params"]["b1"]["min"] == default_info_params["b1"]["min"]
    assert updated_info["params"]["b1"]["max"] == default_info_params["b1"]["max"]


def test_run_file(tmpdir):
    input_file = os.path.join(tmpdir, 'pars.yaml')
    root = os.path.join(tmpdir, 'test')
    yaml_dump_file(input_file, dict(test_info_common, output=root))
    run_script([input_file, '--force'])
    likname = list(test_info_common["likelihood"])[0]
    default_info = get_default_info(likname, "likelihood")
    updated_info = yaml_load_file(root + '.updated.yaml')
    assert updated_info["prior"] == default_info["prior"]



"""
.. module:: theories.camb

:Synopsis: Managing the CAMB cosmological code
:Author: Jesus Torrado and Antony Lewis

.. |br| raw:: html

   <br />

This module imports and manages the CAMB cosmological code.
It requires CAMB 1.1.3 or higher.

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   A. Lewis, A. Challinor, A. Lasenby,
   *Efficient computation of CMB anisotropies in closed FRW*
   (`arXiv:astro-ph/9911177 <https://arxiv.org/abs/astro-ph/9911177>`_)
   |br|
   C. Howlett, A. Lewis, A. Hall, A. Challinor,
   *CMB power spectrum parameter degeneracies in the era of precision cosmology*
   (`arXiv:1201.3654 <https://arxiv.org/abs/1201.3654>`_)


Usage
-----

If you are using a likelihood that requires some observable from CAMB, simply add CAMB
to the theory block.

You can specify any parameter that CAMB understands in the ``params`` block:

.. code-block:: yaml

   theory:
     camb:
       extra_args:
         [any param that CAMB understands, for FIXED and PRECISION]

   params:
       [any param that CAMB understands, fixed, sampled or derived]


If you want to use your own version of CAMB, you need to specify its location with a
``path`` option inside the ``camb`` block. If you do not specify a ``path``,
CAMB will be loaded from the automatic-install ``packages_path`` folder, if specified, or
otherwise imported as a globally-installed Python package. If you want to force that
the global ``camb`` installation is used, pass ``path='global'``. Cobaya will print at
initialisation where CAMB was actually loaded from.


.. _camb_access:

Access to CAMB computation products
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

You can retrieve CAMB computation products within likelihoods (or other pipeline
components in general) or manually from a :class:`~model.Model` as long as you have added
them as requisites; see :doc:`cosmo_external_likelihood` or
:doc:`cosmo_external_likelihood_class` for the likelihood case, and :doc:`cosmo_model` for
the manual case.

The products that you can request and later retrieve are listed in
:func:`~theories.cosmo.BoltzmannBase.must_provide`.

If you would like to access a CAMB result that is not accessible that way, you can access
the full CAMB results object
`CAMBdata <https://camb.readthedocs.io/en/latest/results.html#camb.results.CAMBdata>`_
directly by adding ``{"CAMBdata": None}`` to your requisites, and then retrieving it with
``provider.get_CAMBdata()``.

In general, the use of ``CAMBdata`` should be avoided in public code, since it breaks
compatibility with other Boltzmann codes at the likelihood interface level. If you need
a quantity for a public code that is not generally interfaced in
:func:`~theories.cosmo.BoltzmannBase.must_provide`, let us know if you think it makes
sense to add it.


.. _camb_modify:

Modifying CAMB
^^^^^^^^^^^^^^

If you modify CAMB and add new variables, make sure that the variables you create are
exposed in the Python interface (`instructions here
<https://camb.readthedocs.io/en/latest/model.html#camb.model.CAMBparams>`__).
If you follow those instructions you do not need to make any additional modification in
Cobaya.

If your modification involves new computed quantities, add a retrieving method to
`CAMBdata <https://camb.readthedocs.io/en/latest/results.html#camb.results.CAMBdata>`_,
and see :ref:`camb_access`.

You can use the :doc:`model wrapper <cosmo_model>` to test your modification by
evaluating observables or getting derived quantities at known points in the parameter
space (set ``debug: True`` to get more detailed information of what exactly is passed to
CAMB).

In your CAMB modification, remember that you can raise a ``CAMBParamRangeError`` or a
``CAMBError`` whenever the computation of any observable would fail, but you do not
expect that observable to be compatible with the data (e.g. at the fringes of the
parameter space). Whenever such an error is raised during sampling, the likelihood is
assumed to be zero, and the run is not interrupted.

.. note::

   If your modified CAMB has a lower version number than the minimum required by Cobaya,
   you will get an error at initialisation. You may still be able to use it by setting the
   option ``ignore_obsolete: True`` in the ``camb`` block (though you would be doing that
   at your own risk; ideally you should translate your modification to a newer CAMB
   version, in case there have been important fixes since the release of your baseline
   version).


Installation
------------

Pre-requisites
^^^^^^^^^^^^^^

**cobaya** calls CAMB using its Python interface, which requires that you compile CAMB
using intel's ifort compiler or the GNU gfortran compiler version 6.4 or later.
To check if you have the latter, type ``gfortran --version`` in the shell,
and the first line should look like

.. code::

   GNU Fortran ([your OS version]) [gfortran version] [release date]

Check that ``[gfortran's version]`` is at least 6.4. If you get an error instead, you need
to install gfortran (contact your local IT service).

CAMB comes with binaries pre-built for Windows, so if you don't need to modify the CAMB
source code, no Fortran compiler is needed.

If you are using Anaconda you can also install a pre-compiled CAMB package from conda
forge using

.. code::

  conda install -c conda-forge camb

Automatic installation
^^^^^^^^^^^^^^^^^^^^^^

If you do not plan to modify CAMB, the easiest way to install it is using the
:doc:`automatic installation script <installation_cosmo>`. Just make sure that
``theory: camb:`` appears in one of the files passed as arguments to the installation
script.


Manual installation (or using your own version)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you are planning to modify CAMB or use an already modified version,
you should not use the automatic installation script. Use the installation method that
best adapts to your needs:

* [**Recommended for staying up-to-date**]
  To install CAMB locally and keep it up-to-date, clone the
  `CAMB repository in GitHub <https://github.com/cmbant/CAMB>`_
  in some folder of your choice, say ``/path/to/theories/CAMB``:

  .. code:: bash

      $ cd /path/to/theories
      $ git clone --recursive https://github.com/cmbant/CAMB.git
      $ cd CAMB
      $ python setup.py build

  To update to the last changes in CAMB (master), run ``git pull`` from CAMB's folder and
  re-build using the last command. If you do not want to use multiple versions of CAMB,
  you can also make your local installation available to python generally by installing
  it using

.. code:: bash

     $ python -m pip install -e /path/to/CAMB

* [**Recommended for modifying CAMB**]
  First, `fork the CAMB repository in GitHub <https://github.com/cmbant/CAMB>`_
  (follow `these instructions <https://help.github.com/articles/fork-a-repo/>`_) and then
  follow the same steps as above, substituting the second one with:

  .. code:: bash

      $ git clone --recursive https://[YourGithubUser]@github.com/[YourGithubUser]/CAMB.git

* To use your own version, assuming it's placed under ``/path/to/theories/CAMB``,
  just make sure it is compiled.

In the cases above, you **must** specify the path to your CAMB installation in
the input block for CAMB (otherwise a system-wide CAMB may be used instead):

.. code:: yaml

   theory:
     camb:
       path: /path/to/theories/CAMB

.. note::

   In any of these methods, if you intend to switch between different versions or
   modifications of CAMB you should not install CAMB as python package using
   ``pip install``, as the official instructions suggest. It is not necessary
   if you indicate the path to your preferred installation as explained above.
"""

# Global
import sys
import os
import numbers
import ctypes
import platform
from copy import deepcopy
from typing import NamedTuple, Any, Callable, Optional
import numpy as np
from itertools import chain
# Local
from cobaya.component import ComponentNotInstalledError, load_external_module
from cobaya.theories.cosmo import BoltzmannBase
from cobaya.log import LoggedError, get_logger
from cobaya.install import download_github_release, check_gcc_version
from cobaya.tools import getfullargspec, get_class_methods, get_properties, \
    check_module_version, str_to_list, Pool1D, Pool2D, PoolND, VersionCheckError
from cobaya.theory import HelperTheory
from cobaya.typing import InfoDict, empty_dict


# Result collector
class Collector(NamedTuple):
    method: Callable
    args: list = []
    kwargs: dict = {}
    z_pool: Optional[PoolND] = None
    post: Optional[Callable] = None


class CAMBOutputs(NamedTuple):
    camb_params: Any
    results: Any
    derived: dict


class CAMB(BoltzmannBase):
    r"""
    CAMB cosmological Boltzmann code \cite{Lewis:1999bs,Howlett:2012mh}.
    """

    # Name of the Class repo/folder and version to download
    _camb_repo_name = "cmbant/CAMB"
    _camb_repo_version = os.environ.get("CAMB_REPO_VERSION", "master")
    _camb_min_gcc_version = "6.4"
    _min_camb_version = '1.3.5'

    file_base_name = 'camb'
    external_primordial_pk: bool
    camb: Any
    ignore_obsolete: bool

    def initialize(self):
        """Importing CAMB from the correct path, if given."""
        try:
            install_path = (lambda p: self.get_path(p) if p else None)(self.packages_path)
            min_version = None if self.ignore_obsolete else self._min_camb_version
            self.camb = load_external_module(
                "camb", path=self.path, install_path=install_path,
                min_version=min_version, get_import_path=self.get_import_path,
                logger=self.log, not_installed_level="debug")
        except VersionCheckError as excpt:
            raise VersionCheckError(
                str(excpt) + " If you are using CAMB unmodified, upgrade with"
                             "`cobaya-install camb --upgrade`. If you are using a modified CAMB, "
                             "set the option `ignore_obsolete: True` for CAMB.")
        except ComponentNotInstalledError as excpt:
            raise ComponentNotInstalledError(
                self.log, (f"Could not find CAMB: {excpt}. "
                           "To install it, run `cobaya-install camb`"))
        super().initialize()
        self.extra_attrs = {"Want_CMB": False, "Want_cl_2D_array": False,
                            'WantCls': False}
        # Derived parameters that may not have been requested, but will be necessary later
        self.derived_extra = []
        # Some default settings
        self.needs_perts = False
        self.limber = False
        self.non_linear_sources = False
        self.non_linear_pk = False
        self._base_params = None
        self._needs_lensing_cross = False
        self._sigmaR_z_indices = {}

        if self.external_primordial_pk:
            self.extra_args['initial_power_model'] \
                = self.camb.initialpower.SplinedInitialPower
            self.initial_power_args, self.power_params = {}, []
        else:
            power_spectrum = self.camb.CAMBparams.make_class_named(
                self.extra_args.get('initial_power_model',
                                    self.camb.initialpower.InitialPowerLaw),
                self.camb.initialpower.InitialPower)
            self.initial_power_args, self.power_params = \
                self._extract_params(power_spectrum.set_params)

        nonlin = self.camb.CAMBparams.make_class_named(
            self.extra_args.get('non_linear_model',
                                self.camb.nonlinear.Halofit),
            self.camb.nonlinear.NonLinearModel)

        self.nonlin_args, self.nonlin_params = self._extract_params(nonlin.set_params)

        self.requires = str_to_list(getattr(self, "requires", []))
        self._transfer_requires = [p for p in self.requires if
                                   p not in self.get_can_support_params()]
        self.requires = [p for p in self.requires if p not in self._transfer_requires]

    def _extract_params(self, set_func):
        args = {}
        params = []
        pars = getfullargspec(set_func)
        for arg in pars.args[1:len(pars.args) - len(pars.defaults or [])]:
            params.append(arg)
        if pars.defaults:
            for arg, v in zip(pars.args[len(pars.args) - len(pars.defaults):],
                              pars.defaults):
                if arg in self.extra_args:
                    args[arg] = self.extra_args.pop(arg)
                elif (isinstance(v,
                                 numbers.Number) or v is None) and 'version' not in arg:
                    params.append(arg)
        return args, params

    def initialize_with_params(self):
        # must set WantTensors manually if using external_primordial_pk
        if not self.external_primordial_pk \
                and set(self.input_params).intersection({'r', 'At'}):
            self.extra_attrs["WantTensors"] = True
            self.extra_attrs["Accuracy.AccurateBB"] = True

    def get_can_support_params(self):
        return self.power_params + self.nonlin_params

    def get_allow_agnostic(self):
        return False

    def set_cl_reqs(self, reqs):
        """
        Sets some common settings for both lensed and unlensed Cl's.
        """
        self.extra_args["lmax"] = max(
            max(reqs.values()), self.extra_args.get("lmax", 0))
        self.needs_perts = True
        self.extra_attrs["Want_CMB"] = True
        self.extra_attrs["WantCls"] = True
        if 'TCMB' not in self.derived_extra:
            self.derived_extra += ['TCMB']

    def must_provide(self, **requirements):
        # Computed quantities required by the likelihoods
        # Note that redshifts below are treated differently for background quantities,
        #   where no additional transfer computation is needed (e.g. H(z)),
        #   and matter-power-related quantities, that require additional computation
        #   and need the redshifts to be passed at CAMBParams instantiation.
        #   Also, we always make sure that those redshifts are sorted in descending order,
        #   since all CAMB related functions return quantities in that implicit order
        # The following super call makes sure that the requirements are properly
        # accumulated, i.e. taking the max of precision requests, etc.
        super().must_provide(**requirements)
        CAMBdata = self.camb.CAMBdata

        for k, v in self._must_provide.items():
            # Products and other computations
            if k == "Cl":
                self.set_cl_reqs(v)
                cls = [a.lower() for a in v]
                needs_lensing = set(cls).intersection({"pp", "pt", "pe", "tp", "ep"})
                self.collectors[k] = Collector(
                    method=CAMBdata.get_cmb_power_spectra,
                    kwargs={
                        "spectra": list(set(
                            (self.collectors[k].kwargs.get("spectra", [])
                             if k in self.collectors else []) +
                            ["total"] + (["lens_potential"] if needs_lensing else []))),
                        "raw_cl": False})
                if "pp" in cls and self.extra_args.get(
                        "lens_potential_accuracy") is None:
                    self.extra_args["lens_potential_accuracy"] = 1
                self.non_linear_sources = self.extra_args.get("lens_potential_accuracy",
                                                              1) >= 1
                if set(cls).intersection({"pt", "pe", "tp", "ep"}):
                    self._needs_lensing_cross = True
            elif k == "unlensed_Cl":
                self.set_cl_reqs(v)
                self.collectors[k] = Collector(
                    method=CAMBdata.get_cmb_power_spectra,
                    kwargs={"spectra": ["unlensed_total"], "raw_cl": False})
            elif k == "Hubble":
                self.set_collector_with_z_pool(k, v["z"], CAMBdata.h_of_z)
            elif k in ["Omega_b", "Omega_cdm", "Omega_nu_massive"]:
                varnames = {
                    "Omega_b": "baryon", "Omega_cdm": "cdm", "Omega_nu_massive": "nu"}
                self.set_collector_with_z_pool(
                    k, v["z"], CAMBdata.get_Omega, kwargs={"var": varnames[k]})
            elif k in ("angular_diameter_distance", "comoving_radial_distance"):
                self.set_collector_with_z_pool(k, v["z"], getattr(CAMBdata, k))
            elif k == "angular_diameter_distance_2":
                check_module_version(self.camb, '1.3.5')
                self.set_collector_with_z_pool(
                    k, v["z_pairs"], CAMBdata.angular_diameter_distance2, d=2)
            elif k == "sigma8_z":
                self.add_to_redshifts(v["z"])
                self.collectors[k] = Collector(
                    method=CAMBdata.get_sigma8,
                    kwargs={},
                    post=(lambda *x: x[::-1]))  # returned in inverse order
                self.needs_perts = True
            elif k == "fsigma8":
                self.add_to_redshifts(v["z"])
                self.collectors[k] = Collector(
                    method=CAMBdata.get_fsigma8,
                    kwargs={},
                    post=(lambda *x: x[::-1]))  # returned in inverse order
                self.needs_perts = True
            elif isinstance(k, tuple) and k[0] == "sigma_R":
                kwargs = v.copy()
                self.extra_args["kmax"] = max(kwargs.pop("k_max"),
                                              self.extra_args.get("kmax", 0))
                redshifts = kwargs.pop("z")
                self.add_to_redshifts(redshifts)
                var_pair = k[1:]

                def get_sigmaR(results, **tmp):
                    _indices = self._sigmaR_z_indices.get(var_pair)
                    if _indices is None or list(_indices) == []:
                        z_indices = []
                        calc = np.array(results.Params.Transfer.PK_redshifts[
                                        :results.Params.Transfer.PK_num_redshifts])
                        for z in redshifts:
                            for i, zcalc in enumerate(calc):
                                if np.isclose(zcalc, z, rtol=1e-4):
                                    z_indices += [i]
                                    break
                            else:
                                raise LoggedError(self.log, "sigma_R redshift not found"
                                                            "in computed P_K array %s", z)
                        _indices = np.array(z_indices, dtype=np.int32)
                        self._sigmaR_z_indices[var_pair] = _indices
                    R, z, sigma = results.get_sigmaR(hubble_units=False, return_R_z=True,
                                                     z_indices=_indices, **tmp)
                    return z, R, sigma

                kwargs.update(dict(zip(["var1", "var2"], var_pair)))
                self.collectors[k] = Collector(method=get_sigmaR, kwargs=kwargs)
                self.needs_perts = True
            elif isinstance(k, tuple) and k[0] == "Pk_grid":
                kwargs = v.copy()
                self.extra_args["kmax"] = max(kwargs.pop("k_max"),
                                              self.extra_args.get("kmax", 0))
                self.add_to_redshifts(kwargs.pop("z"))
                # need to ensure can't have conflicts between requests from
                # different likelihoods. Store results without Hubble units.
                if kwargs.get("hubble_units", False) or kwargs.get("k_hunit", False):
                    raise LoggedError(self.log, "hubble_units and k_hunit must be False"
                                                "for consistency")
                kwargs["hubble_units"] = False
                kwargs["k_hunit"] = False
                if kwargs["nonlinear"]:
                    self.non_linear_pk = True
                var_pair = k[2:]
                kwargs.update(dict(zip(["var1", "var2"], var_pair)))
                self.collectors[k] = Collector(
                    method=CAMBdata.get_linear_matter_power_spectrum,
                    kwargs=kwargs.copy())
                self.needs_perts = True
            elif k == "source_Cl":
                if not getattr(self, "sources", None):
                    self.sources: InfoDict = {}
                for source, window in v["sources"].items():
                    # If it was already there, BoltzmannBase.must_provide() has already
                    # checked that old info == new info
                    if source not in self.sources:
                        self.sources[source] = window
                self.limber = v.get("limber", True)
                self.non_linear_sources = self.non_linear_sources or \
                                          v.get("non_linear", False)
                if "lmax" in v:
                    self.extra_args["lmax"] = max(v["lmax"],
                                                  self.extra_args.get("lmax", 0))
                self.needs_perts = True
                self.collectors[k] = Collector(method=CAMBdata.get_source_cls_dict)
                self.extra_attrs["Want_cl_2D_array"] = True
                self.extra_attrs["WantCls"] = True
            elif k == 'CAMBdata':
                # Just get CAMB results object
                self.collectors[k] = None
            elif v is None:
                # General derived parameters
                k_translated = self.translate_param(k)
                if k_translated not in self.derived_extra:
                    self.derived_extra += [k_translated]
                if k == "sigma8":
                    self.extra_attrs["WantTransfer"] = True
                    self.needs_perts = True
                    self.add_to_redshifts([0.])
            else:
                raise LoggedError(self.log, "This should not be happening. Contact the "
                                            "developers.")
        self.check_no_repeated_input_extra()

        # Computing non-linear corrections
        model = self.camb.model
        self.extra_attrs["NonLinear"] = {
            (True, True): model.NonLinear_both,
            (True, False): model.NonLinear_lens,
            (False, True): model.NonLinear_pk,
            (False, False): False}[(self.non_linear_sources, self.non_linear_pk)]
        # set-set base CAMB params if anything might have changed
        self._base_params = None

        must_provide: InfoDict = {
            'CAMB_transfers': {'non_linear': self.non_linear_sources,
                               'needs_perts': self.needs_perts}}
        if self.external_primordial_pk and self.needs_perts:
            must_provide['primordial_scalar_pk'] = {'lmax': self.extra_args.get("lmax"),
                                                    'kmax': self.extra_args.get('kmax')}
            if self.extra_attrs.get('WantTensors'):
                must_provide['primordial_tensor_pk'] = {'lmax':
                    self.extra_attrs.get(
                        "max_l_tensor", self.extra_args.get("lmax"))}
        return must_provide

    def add_to_redshifts(self, z):
        """
        Adds redshifts to the list of them for which CAMB computes perturbations.
        """
        if not hasattr(self, "z_pool_for_perturbations"):
            self.z_pool_for_perturbations = Pool1D(z)
        else:
            self.z_pool_for_perturbations.update(z)
        self.extra_args["redshifts"] = np.flip(self.z_pool_for_perturbations.values)

    def set_collector_with_z_pool(self, k, zs, method, args=(), kwargs=empty_dict, d=1):
        """
        Creates a collector for a z-dependent quantity, keeping track of the pool of z's.
        """
        if k in self.collectors:
            z_pool = self.collectors[k].z_pool
            z_pool.update(zs)
        else:
            Pool = {1: Pool1D, 2: Pool2D}[d]
            z_pool = Pool(zs)
        if d == 1:
            kwargs_with_z = {"z": z_pool.values}
        else:
            kwargs_with_z = {"z1": np.array(z_pool.values[:, 0]),
                             "z2": np.array(z_pool.values[:, 1])}
        kwargs_with_z.update(kwargs)
        self.collectors[k] = Collector(
            method=method, z_pool=z_pool, kwargs=kwargs_with_z, args=args)

    def calculate(self, state, want_derived=True, **params_values_dict):
        try:
            params, results = self.provider.get_CAMB_transfers()
            if self.collectors or 'sigma8' in self.derived_extra:
                if self.external_primordial_pk and self.needs_perts:
                    primordial_pk = self.provider.get_primordial_scalar_pk()
                    if primordial_pk.get('log_regular', True):
                        results.Params.InitPower.set_scalar_log_regular(
                            primordial_pk['kmin'], primordial_pk['kmax'],
                            primordial_pk['Pk'])
                    else:
                        results.Params.InitPower.set_scalar_table(
                            primordial_pk['k'], primordial_pk['Pk']
                        )
                    results.Params.InitPower.effective_ns_for_nonlinear = \
                        primordial_pk.get('effective_ns_for_nonlinear', 0.97)
                    if self.extra_attrs.get("WantTensors"):
                        primordial_pk = self.provider.get_primordial_tensor_pk()
                        if primordial_pk.get('log_regular', True):
                            results.Params.InitPower.set_tensor_log_regular(
                                primordial_pk['kmin'], primordial_pk['kmax'],
                                primordial_pk['Pk'])
                        else:
                            results.Params.InitPower.set_tensor_table(
                                primordial_pk['k'], primordial_pk['Pk']
                            )
                else:
                    args = {self.translate_param(p): v for p, v in
                            params_values_dict.items() if p in self.power_params}
                    args.update(self.initial_power_args)
                    results.Params.InitPower.set_params(**args)
                if self.non_linear_sources or self.non_linear_pk:
                    args = {self.translate_param(p): v for p, v in
                            params_values_dict.items() if p in self.nonlin_params}
                    args.update(self.nonlin_args)
                    results.Params.NonLinearModel.set_params(**args)
                results.power_spectra_from_transfer()
            for product, collector in self.collectors.items():
                if collector:
                    state[product] = \
                        collector.method(results, *collector.args, **collector.kwargs)
                    if collector.post:
                        state[product] = collector.post(*state[product])
                else:
                    state[product] = results
        except self.camb.baseconfig.CAMBError as e:
            if self.stop_at_error:
                self.log.error(
                    "Computation error (see traceback below)! "
                    "Parameters sent to CAMB: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    dict(state["params"]), dict(self.extra_args))
                raise
            else:
                # Assumed to be a "parameter out of range" error.
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on. "
                               "The output of the CAMB error was %s" % e)
                return False
            # Prepare derived parameters
        intermediates = CAMBOutputs(params, results,
                                    results.get_derived_params() if results else None)
        if want_derived:
            state["derived"] = self._get_derived_output(intermediates)
        # Prepare necessary extra derived parameters
        state["derived_extra"] = {
            p: self._get_derived(p, intermediates) for p in self.derived_extra}

    @staticmethod
    def _get_derived(p, intermediates):
        """
        General function to extract a single derived parameter.

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        if intermediates.derived:
            derived = intermediates.derived.get(p, None)
            if derived is not None:
                return derived
        # Specific calls, if general ones fail:
        if p == "sigma8":
            return intermediates.results.get_sigma8_0()
        try:
            return getattr(intermediates.camb_params, p)
        except AttributeError:
            try:
                return getattr(intermediates.results, p)
            except AttributeError:
                return getattr(intermediates.camb_params, "get_" + p, lambda: None)()

    def _get_derived_output(self, intermediates):
        """
        Returns a dictionary of derived parameters with their values,
        using the *current* state (i.e. it should only be called from
        the ``compute`` method).

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        derived = {}
        for p in self.output_params:
            derived[p] = self._get_derived(self.translate_param(p), intermediates)
            if derived[p] is None:
                raise LoggedError(self.log, "Derived param '%s' not implemented"
                                            " in the CAMB interface", p)
        return derived

    def _get_Cl(self, ell_factor=False, units="FIRASmuK2", lensed=True):
        which_key = "Cl" if lensed else "unlensed_Cl"
        which_result = "total" if lensed else "unlensed_total"
        which_error = "lensed" if lensed else "unlensed"
        try:
            cl_camb = self.current_state[which_key][which_result].copy()
        except:
            raise LoggedError(self.log, "No %s Cl's were computed. Are you sure that you "
                                        "have requested them?", which_error)
        units_factor = self._cmb_unit_factor(
            units, self.current_state['derived_extra']['TCMB'])
        ls = np.arange(cl_camb.shape[0], dtype=np.int64)
        if not ell_factor:
            # unit conversion and ell_factor. CAMB output is *with* the factors already
            ells_factor = ls[1:] * (ls[1:] + 1)
            cl_camb[1:, :] /= ells_factor[..., np.newaxis]
            cl_camb[1:, :] *= (2 * np.pi) * units_factor ** 2
        elif units_factor != 1:
            cl_camb *= units_factor ** 2
        mapping = {"tt": 0, "ee": 1, "bb": 2, "te": 3, "et": 3}
        cls = {"ell": ls}
        for sp, i in mapping.items():
            cls[sp] = cl_camb[:, i]
        if lensed:
            cl_lens: Optional[np.ndarray] = self.current_state["Cl"].get("lens_potential")
            if cl_lens is not None:
                cls["pp"] = cl_lens[:, 0].copy()
                if not ell_factor:
                    # noinspection PyUnboundLocalVariable
                    cls["pp"][1:] /= ells_factor ** 2 / (2 * np.pi)
                if self._needs_lensing_cross:
                    for i, cross in enumerate(['pt', 'pe']):
                        cls[cross] = cl_lens[:, i + 1].copy() * units_factor
                        if not ell_factor:
                            cls[cross][1:] /= ells_factor ** (3. / 2) / (2 * np.pi)
                        cls[cross[::-1]] = cls[cross]
        return cls

    def get_Cl(self, ell_factor=False, units="FIRASmuK2"):
        return self._get_Cl(ell_factor=ell_factor, units=units, lensed=True)

    def get_unlensed_Cl(self, ell_factor=False, units="FIRASmuK2"):
        return self._get_Cl(ell_factor=ell_factor, units=units, lensed=False)

    def _get_z_dependent(self, quantity, z, _pool=None):
        # Partially reimplemented because of sigma8_z, etc, use different pool
        pool = None
        if quantity in ["sigma8_z", "fsigma8"]:
            pool = self.z_pool_for_perturbations
        return super()._get_z_dependent(quantity, z, pool=pool)

    def get_source_Cl(self):
        # get C_l^XX from the cosmological code
        try:
            cls = deepcopy(self.current_state["source_Cl"])
        except:
            raise LoggedError(
                self.log, "No source Cl's were computed. "
                          "Are you sure that you have requested some source?")
        cls_dict: dict = dict()
        for term, cl in cls.items():
            term_tuple = tuple(
                (lambda x: x if x == "P" else list(self.sources)[int(x) - 1])(
                    _.strip("W")) for _ in term.split("x"))
            cls_dict[term_tuple] = cl
        cls_dict["ell"] = np.arange(cls[list(cls)[0]].shape[0])
        return cls_dict

    def get_CAMBdata(self):
        """
        Get the CAMB result object (must have been requested as a requirement).

        :return: CAMB's `CAMBdata <https://camb.readthedocs.io/en/latest/results.html>`_
                 result instance for the current parameters
        """
        return self.current_state['CAMBdata']

    def get_can_provide_params(self):
        # possible derived parameters for derived_extra, excluding things that are
        # only input parameters.
        params_derived = list(get_class_methods(self.camb.CAMBparams))
        params_derived.remove("custom_source_names")
        fields = []
        # noinspection PyProtectedMember
        for f, tp in self.camb.CAMBparams._fields_:
            if tp is ctypes.c_double and 'max_eta_k' not in f \
                    and f not in ['Alens', 'num_nu_massless']:
                fields.append(f)
        fields += ['omega_de', 'sigma8']  # only parameters from CAMBdata
        properties = get_properties(self.camb.CAMBparams)
        names = self.camb.model.derived_names + properties + fields + params_derived
        for name, mapped in self.renames.items():
            if mapped in names:
                names.append(name)
        # remove any parameters explicitly tagged as input requirements
        return set(names).difference(chain(self._transfer_requires, self.requires))

    def get_version(self):
        return self.camb.__version__

    def set(self, params_values_dict, state):
        # Prepare parameters to be passed: this is called from the CambTransfers instance
        args = {self.translate_param(p): v for p, v in params_values_dict.items()}
        # Generate and save
        self.log.debug("Setting parameters: %r and %r",
                       dict(args), dict(self.extra_args))
        try:
            if not self._base_params:
                base_args = args.copy()
                base_args.update(self.extra_args)
                # Remove extra args that might
                # cause an error if the associated product is not requested
                if not self.extra_attrs["WantCls"]:
                    for not_needed in getfullargspec(
                            self.camb.CAMBparams.set_for_lmax).args[1:]:
                        base_args.pop(not_needed, None)
                self._reduced_extra_args = self.extra_args.copy()
                params = self.camb.set_params(**base_args)
                # pre-set the parameters that are not varying
                for non_param_func in ['set_classes', 'set_matter_power', 'set_for_lmax']:
                    for fixed_param in getfullargspec(
                            getattr(self.camb.CAMBparams, non_param_func)).args[1:]:
                        if fixed_param in args:
                            raise LoggedError(
                                self.log, "Trying to sample fixed theory parameter %s",
                                fixed_param)
                        self._reduced_extra_args.pop(fixed_param, None)
                if self.extra_attrs:
                    self.log.debug("Setting attributes of CAMBparams: %r",
                                   self.extra_attrs)
                for attr, value in self.extra_attrs.items():
                    obj = params
                    if '.' in attr:
                        parts = attr.split('.')
                        for p in parts[:-1]:
                            obj = getattr(obj, p)
                        par = parts[-1]
                    else:
                        par = attr
                    if hasattr(obj, par):
                        setattr(obj, par, value)
                    else:
                        raise LoggedError(
                            self.log,
                            "Some of the attributes to be set manually were not "
                            "recognized: %s=%s", attr, value)
                # Sources
                if getattr(self, "sources", None):
                    self.log.debug("Setting sources: %r", self.sources)
                    sources = self.camb.sources
                    source_windows = []
                    for source, window in self.sources.items():
                        function = window.pop("function", None)
                        if function == "spline":
                            source_windows.append(sources.SplinedSourceWindow(**window))
                        elif function == "gaussian":
                            source_windows.append(sources.GaussianSourceWindow(**window))
                        else:
                            raise LoggedError(self.log,
                                              "Unknown source window function type %r",
                                              function)
                        window["function"] = function
                    params.SourceWindows = source_windows
                    params.SourceTerms.limber_windows = self.limber
                self._base_params = params
            args.update(self._reduced_extra_args)
            return self.camb.set_params(self._base_params.copy(), **args)
        except self.camb.baseconfig.CAMBParamRangeError:
            if self.stop_at_error:
                raise LoggedError(self.log, "Out of bound parameters: %r",
                                  params_values_dict)
            else:
                self.log.debug("Out of bounds parameters. "
                               "Assigning 0 likelihood and going on.")
        except (self.camb.baseconfig.CAMBValueError, self.camb.baseconfig.CAMBError):
            if self.stop_at_error:
                self.log.error(
                    "Error setting parameters (see traceback below)! "
                    "Parameters sent to CAMB: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    dict(state["params"]), dict(self.extra_args))
                raise
        except self.camb.baseconfig.CAMBUnknownArgumentError as e:
            raise LoggedError(
                self.log,
                "Some of the parameters passed to CAMB were not recognized: %s" % str(e))
        return False

    def get_helper_theories(self):
        """
        Transfer functions are computed separately by camb.transfers, then this
        class uses the transfer functions to calculate power spectra (using A_s, n_s etc).
        """
        self._camb_transfers = CambTransfers(self, 'camb.transfers',
                                             dict(stop_at_error=self.stop_at_error),
                                             timing=self.timer)
        setattr(self._camb_transfers, "requires", self._transfer_requires)
        return {'camb.transfers': self._camb_transfers}

    def get_speed(self):
        if self._measured_speed:
            return self._measured_speed
        if not self.non_linear_sources:
            return self.speed * 10
        if {'omk', 'omegak'}.intersection(set(self._camb_transfers.input_params)):
            return self.speed / 1.5
        return self.speed * 3

    @classmethod
    def get_path(cls, path):
        return os.path.realpath(
            os.path.join(path, "code",
                         cls._camb_repo_name[cls._camb_repo_name.find("/") + 1:]))

    @staticmethod
    def get_import_path(path):
        """
        Returns the ``camb`` module import path if there is a compiled version of CAMB in
        the given folder. Otherwise raises ``FileNotFoundError``.
        """
        lib_fname = "cambdll.dll" if platform.system() == "Windows" else "camblib.so"
        if not os.path.isfile(os.path.realpath(os.path.join(path, "camb", lib_fname))):
            raise FileNotFoundError(
                f"Could not find compiled CAMB library {lib_fname} in {path}.")
        return path

    @classmethod
    def is_installed(cls, reload=False, **kwargs):
        if not kwargs.get("code", True):
            return True
        try:
            return bool(load_external_module(
                "camb", path=kwargs["path"], get_import_path=cls.get_import_path,
                min_version=cls._min_camb_version, reload=reload,
                logger=get_logger(cls.__name__), not_installed_level="debug"))
        except ComponentNotInstalledError:
            return False

    @classmethod
    def install(cls, path=None, code=True, no_progress_bars=False, **_kwargs):
        log = get_logger(cls.__name__)
        if not code:
            log.info("Code not requested. Nothing to do.")
            return True
        log.info("Downloading camb...")
        success = download_github_release(
            os.path.join(path, "code"), cls._camb_repo_name, cls._camb_repo_version,
            no_progress_bars=no_progress_bars, logger=log)
        if not success:
            log.error("Could not download camb.")
            return False
        camb_path = cls.get_path(path)
        log.info("Compiling camb...")
        from subprocess import Popen, PIPE
        process_make = Popen([sys.executable, "setup.py", "build_cluster"],
                             cwd=camb_path, stdout=PIPE, stderr=PIPE)
        out, err = process_make.communicate()
        if process_make.returncode:
            log.info(out.decode())
            log.info(err.decode())
            gcc_check = check_gcc_version(cls._camb_min_gcc_version, error_returns=False)
            if not gcc_check:
                cause = (" Possible cause: it looks like `gcc` does not have the correct "
                         "version number (CAMB requires %s); and `ifort` is also "
                         "probably not available." % cls._camb_min_gcc_version)
            else:
                cause = ""
            log.error("Compilation failed!" + cause)
            return False
        return True


class CambTransfers(HelperTheory):
    """
    Helper theory class that calculates transfer functions only. The result is cached
    when only initial power spectrum or non-linear model parameters change
    """

    def __init__(self, cobaya_camb, name, info, timing=None):
        self.needs_perts = False
        self.non_linear_sources = False
        super().__init__(info, name, timing=timing)
        self.cobaya_camb = cobaya_camb
        self.camb = cobaya_camb.camb
        self.speed = self.cobaya_camb.speed * 1.5

    def get_can_support_params(self):
        supported_params = self.camb.get_valid_numerical_params(
            transfer_only=True,
            dark_energy_model=self.cobaya_camb.extra_args.get('dark_energy_model'),
            recombination_model=self.cobaya_camb.extra_args.get('recombination_model')) \
                           - set(self.cobaya_camb.extra_args) \
                           - set(self.cobaya_camb.extra_attrs)

        for name, mapped in self.cobaya_camb.renames.items():
            if mapped in supported_params:
                supported_params.add(name)
        return supported_params

    def get_allow_agnostic(self):
        return False

    def must_provide(self, **requirements):
        super().must_provide(**requirements)
        opts = requirements.get('CAMB_transfers')
        if opts:
            self.non_linear_sources = opts['non_linear']
            self.needs_perts = opts['needs_perts']
        self.cobaya_camb.check_no_repeated_input_extra()

    def get_CAMB_transfers(self):
        return self.current_state['results']

    def calculate(self, state, want_derived=True, **params_values_dict):
        # Set parameters
        camb_params = self.cobaya_camb.set(params_values_dict, state)
        # Failed to set parameters but no error raised
        # (e.g. out of computationally feasible range): lik=0
        if not camb_params:
            return False
        # Compute the transfer functions
        try:
            if self.non_linear_sources:
                # only need time sources if non-linear lensing or other non-linear
                # sources. Not needed just for non-linear PK.
                results = self.camb.get_transfer_functions(camb_params,
                                                           only_time_sources=True)
            else:
                results = self.camb.get_transfer_functions(camb_params) \
                    if self.needs_perts else self.camb.get_background(camb_params)
            state['results'] = (camb_params, results)
        except self.camb.baseconfig.CAMBError as e:
            if self.stop_at_error:
                self.log.error(
                    "Computation error (see traceback below)! "
                    "Parameters sent to CAMB: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    dict(state["params"]), dict(self.cobaya_camb.extra_args))
                raise
            else:
                # Assumed to be a "parameter out of range" error.
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on. "
                               "The output of the CAMB error was %s" % e)
                return False

    def initialize_with_params(self):
        if len(set(self.input_params).intersection(
                {"H0", "cosmomc_theta", "thetastar"})) > 1:
            raise LoggedError(self.log, "Can't pass more than one of H0, "
                                        "theta, cosmomc_theta to CAMB.")
        if len(set(self.input_params).intersection({"tau", "zrei"})) > 1:
            raise LoggedError(self.log, "Can't pass more than one of tau and zrei "
                                        "to CAMB.")

        super().initialize_with_params()



# Tries to evaluate the BK18 likelihood at a reference point

from copy import deepcopy
from .common_cosmo import body_of_test
from cobaya.cosmo_input import planck_precision

camb_extra = deepcopy(planck_precision["camb"])
camb_extra.update({"halofit_version": "takahashi"})
classy_extra = deepcopy(planck_precision["classy"])
classy_extra.update({"non linear": "halofit"})
classy_extra.update({"nonlinear_min_k_max": 20})


def test_bicep_keck_2018_camb(packages_path, skip_not_installed):
    info_theory = {"camb": {"extra_args": camb_extra}}
    body_of_test(packages_path, test_point, lik_info, info_theory, chi2,
                 extra_model={"primordial": "SFSR_t"},
                 skip_not_installed=skip_not_installed)


def test_bicep_keck_2018_classy(packages_path, skip_not_installed):
    info_theory = {"classy": {"extra_args": classy_extra}}
    # extra tolerance for CLASS
    body_of_test(packages_path, test_point, lik_info, info_theory, chi2,
                 extra_model={"primordial": "SFSR_t"},
                 skip_not_installed=skip_not_installed)


lik_info = {"bicep_keck_2018": {}}

# NB: chi2 obtained using CAMB w HMcode
chi2 = {"bicep_keck_2018": 543.25, "tolerance": 0.16}

test_point = {
    "omegabh2": 0.2235620E-01,
    "omegach2": 0.1204042E+00,
    "theta_MC_100": 0.1040871E+01,  # for CAMB
    "H0": 0.6718506E+02,  # for CLASS
    "tau": 0.5454114E-01,
    "logA": 0.3046322E+01,
    "ns": 0.9654113E+00,
    "r": 0.1451578E-01,
    "calPlanck": 0.1000689E+01,
    "BBdust": 0.4648994E+01,
    "BBsync": 0.1542620E+01,
    "BBalphadust": -0.5338430E+00,
    "BBbetadust": 0.1576173E+01,
    "BBalphasync": -0.1915241E+00,
    "BBbetasync": -0.3040606E+01,
    "BBdustsynccorr": -0.3441905E+00
}



from cobaya.likelihoods.base_classes import Planck2018CamSpecPython


class TTTEEE(Planck2018CamSpecPython):
    r"""
    High-$\ell$ temperature and polarization \textsc{CamSpec} likelihood of Planck's 2018
    data release \cite{Aghanim:2019ame} (native Python re-implemetation, by A.~Lewis).
    """
    pass



from cobaya.likelihoods.base_classes import BAO


class sdss_dr7_mgs(BAO):
    r"""
    Likelihood of the BAO scale measurement of SDSS DR7 Main Galaxy Sample
    \cite{Ross:2014qpa}.
    """
    pass



from cobaya.likelihoods.base_classes import Planck2018Clik


class TT_clik(Planck2018Clik):
    r"""
    Low-$\ell$ temperature-only likelihood of Planck's 2018 data release
    \cite{Aghanim:2019ame}.
    """
    pass



from cobaya.likelihoods.base_classes import SN


class jla(SN):
    r"""
    Likelihood of the JLA type Ia supernova sample \cite{Betoule:2014frx}, based on
    observations obtained by the SDSS-II and SNLS collaborations.
    """
    pass



from io import StringIO
import os
from cobaya.doc import doc_script
from cobaya.bib import bib_script
from .common import stdout_redirector


def test_doc():
    stream = StringIO()
    with stdout_redirector(stream):
        doc_script([])
    assert 'planck_2018_lensing' in stream.getvalue()
    stream = StringIO()
    with stdout_redirector(stream):
        doc_script(['des_y1.shear'])
    assert 'DES_DzS1' in stream.getvalue()

    stream = StringIO()
    with stdout_redirector(stream):
        doc_script(['planck_2018_highl_CamSpec.TT'])
    assert ' [../base_classes/planck_calib, params_TT_CamSpec]' in stream.getvalue()
    stream = StringIO()
    with stdout_redirector(stream):
        doc_script(['planck_2018_highl_CamSpec.TT', '--expand'])
    assert 'lambda aksz, asz143' in stream.getvalue()


yaml = """
likelihood:
 des_y1.clustering:
 bao.generic:
 planck_2018_lensing.clik:
theory:
 camb:
sampler:
 minimize:
"""


def test_bib(tmpdir):
    stream = StringIO()
    with stdout_redirector(stream):
        bib_script(['des_y1.shear', 'camb', 'mcmc'])
    assert 'Neal:2005' in stream.getvalue()

    f = os.path.join(tmpdir, 'input.yaml')
    with open(f, 'w', encoding='utf-8') as output:
        output.write(yaml)

    stream = StringIO()
    with stdout_redirector(stream):
        bib_script([f])
    assert 'Torrado:2020' in stream.getvalue()



import subprocess
from distutils import spawn
from . import batchjob_args, jobqueue


def delete_jobs():
    Opts = batchjob_args.BatchArgs('Delete running or queued jobs', importance=True,
                                   batchPathOptional=True)

    group = Opts.parser.add_mutually_exclusive_group()
    group.add_argument('--queued', action='store_true')
    group.add_argument('--running', action='store_true')

    Opts.parser.add_argument('--delete_id_min', type=int)
    Opts.parser.add_argument('--delete_id_range', nargs=2, type=int)
    Opts.parser.add_argument('--delete_ids', nargs='+', type=int)
    Opts.parser.add_argument('--confirm', action='store_true')

    (batch, args) = Opts.parseForBatch()

    if batch:
        if args.delete_id_range is not None:
            jobqueue.deleteJobs(args.batchPath, jobId_minmax=args.delete_id_range,
                                confirm=args.confirm)
        if args.delete_id_min is not None:
            jobqueue.deleteJobs(args.batchPath, jobId_min=args.delete_id_min,
                                confirm=args.confirm)
        elif args.delete_ids is not None:
            jobqueue.deleteJobs(args.batchPath, args.delete_ids, confirm=args.confirm)
        else:
            items = [jobItem for jobItem in Opts.filteredBatchItems()]
            batchNames = set(
                [jobItem.name for jobItem in items] + [jobItem.name + '_minimize' for
                                                       jobItem in items])
            jobqueue.deleteJobs(args.batchPath, rootNames=batchNames,
                                confirm=args.confirm)

        if not args.confirm:
            print('jobs not actually deleted: add --confirm to really cancel them')

    else:
        ids = []
        if args.delete_id_range is not None:
            ids = list(range(args.delete_id_range[0], args.delete_id_range[1] + 1))
        elif args.delete_ids is not None:
            ids += args.delete_ids
        elif args.name is not None:
            jobqueue.deleteJobs(args.batchPath, rootNames=args.name)
            return
        else:
            print('Must give --delete_id_range, --delete_ids or --name '
                  'if no batch directory')
        for engine in jobqueue.grid_engine_defaults:
            qdel = jobqueue.engine_default(engine, 'qdel')
            if spawn.find_executable(qdel) is not None:
                for jobId in ids:
                    subprocess.check_output(qdel + ' ' + str(jobId), shell=True)
                break


if __name__ == "__main__":
    delete_jobs()



from cobaya.likelihoods.planck_2018_highl_CamSpec2021.planck_2018_CamSpec2021_python import \
    Planck2018CamSpec2021Python


class Planck2020CamSpecPython(Planck2018CamSpec2021Python):
    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1", "asset": "CamSpec_NPIPE.zip",
                       "directory": "planck_NPIPE_CamSpec"}

    bibtex_file = 'CamSpec_NPIPE_2022.bibtex'



"""
Automatic tests of models.rst example.

For not it only checks that it does not fail.
"""

import os

tests_folder = os.path.dirname(os.path.realpath(__file__))
docs_folder = os.path.join(tests_folder, "..", "docs")
docs_src_folder = os.path.join(docs_folder, "src_examples", "advanced_and_models")


def test_docs_model():
    cwd = os.getcwd()
    try:
        os.chdir(docs_src_folder)
        globals_example = {}
        exec(open(os.path.join(docs_src_folder, "model_create.py")).read(),
             globals_example)
    finally:
        # Back to the working directory of the tests, just in case
        os.chdir(cwd)
